#!/usr/bin/env python
from __future__ import print_function
import time
import unittest
from random import randint
import inspect
from ruffus.file_name_parameters import needs_update_check_modify_time, check_input_files_exist, \
    args_param_factory, open_job_history
from ruffus.file_name_parameters import non_str_sequence
from ruffus import output_from, add_inputs
from ruffus import task, combine
import sys

"""

    test_file_name_parameters.py

"""


import os
tempdir = os.path.abspath(os.path.abspath(os.path.splitext(__file__)[0]))
exe_path = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Unit Tests


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

# print ruffus.__version__
history_file = ':memory:'
history_file = None

# use simplejson in place of json for python < 2.6
try:
    import json
except ImportError:
    import simplejson
    json = simplejson


dumps = json.dumps


def touch(filename):
    with open(filename, "w"):
        pass


# =========================================================================================
#   args_param_factory
# =========================================================================================


def lineno():
    """Returns the current line number in our program."""
    return inspect.currentframe().f_back.f_lineno


def double_parameters(params):
    return [(p, p) for p in params]


class Test_args_param_factory(unittest.TestCase):

    #       self.assertEqual(self.seq, range(10))
    #       self.assertTrue(element in self.seq)
    #       self.assertRaises(ValueError, random.sample, self.seq, 20)

    def forwarded_function(self, params):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        it = args_param_factory(params)
        return list(it(None))

    def test_single_job_per_task(self):
        """
        test convenience form for single job per task
        """
        self.assertEqual(self.forwarded_function([["file.input", "file.output", "other", 1]]),
                         double_parameters([['file.input', 'file.output', 'other', 1]]))

    def test_multiple_jobs_per_task(self):
        """
        test normal form for multiple job per task
        """
        params = [
            ["file0input", "file0.output", "other", 1],
            ["file1input", "file1.output", "other", 1],
            ["file2input", "file2.output", "other", 1],
        ]
        self.assertEqual(self.forwarded_function(params),
                         double_parameters(params))

    def test_nested_multiple_jobs_per_task(self):
        """
        test normal form for multiple job per task
        """
        params = [
            [[["file0input"]], "file0.output", "other", 1],
            ["file1input", "file1.output", "other", 1],
            ["file2input", "file2.output", "other", 1],
        ]
        self.assertEqual(self.forwarded_function(params),
                         double_parameters(params))


# =========================================================================================

#   files_re_param_factory

# =========================================================================================

def recursive_replace(p, from_s, to_s):
    """
    recursively replaces file name specifications using regular expressions
    Non-strings are left alone
    """
    if isinstance(p, str):
        return p.replace(from_s, to_s)
    elif non_str_sequence(p):
        return type(p)(recursive_replace(pp, from_s, to_s) for pp in p)
    else:
        return p


def list_generator_factory(list):
    def list_generator(ignored_args):
        for i in list:
            yield i, i
    return list_generator


l1 = [["input1", "output1.test"], [3, "output2.test"], [
    "input3", "output3.test"], [3, ["output4.test", "output.ignored"]], [], [4, 5]]
l2 = [["output4.test", "output.ignored"]]
l3 = []
l4 = [[1, (2, "output5.test")]]
t1 = task.Task(list_generator_factory, "module.func1")
t1.param_generator_func = list_generator_factory(l1)
t2 = task.Task(list_generator_factory, "module.func2")
t2.param_generator_func = list_generator_factory(l2)
t2._is_single_job_single_output = t2._single_job_single_output
t3 = task.Task(list_generator_factory, "module.func3")
t3.param_generator_func = list_generator_factory(l3)
t4 = task.Task(list_generator_factory, "module.func4")
t4.param_generator_func = list_generator_factory(l4)
t4._is_single_job_single_output = t4._single_job_single_output
t5 = task.Task(list_generator_factory, "module.func5")
t5.param_generator_func = None

next_task_id = 1


class Test_files_re_param_factory(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(tempdir):
            os.makedirs(tempdir)
        touch("%s/f%d.output" % (tempdir, 0))
        for i in range(3):
            touch("%s/f%d.test" % (tempdir, i))
        time.sleep(0.1)
        touch("%s/f%d.output" % (tempdir, 1))
        touch("%s/f%d.output" % (tempdir, 2))
        self.tasks = [t1, t2, t3, t4, t5]

    def tearDown(self):
        for i in range(3):
            os.unlink("%s/f%d.test" % (tempdir, i))
        for i in range(3):
            os.unlink("%s/f%d.output" % (tempdir, i))
        os.removedirs(tempdir)
        pass

    #       self.assertEqual(self.seq, range(10))
    #       self.assertTrue(element in self.seq)
    #       self.assertRaises(ValueError, random.sample, self.seq, 20)

    # def get_param_iterator (self, *old_args):

    def get_param_iterator(self, *unnamed_args, **named_args):
        #
        # replace function / function names with tasks
        #
        # fake virgin task
        # use global incrementing index to avoid name clashes
        #fake_task = task.Task("module", "func_fake%d" % randint(1, 1000000))
        global next_task_id
        next_task_id += 1
        fake_task = task.Task(list_generator_factory,
                              "module.func_fake%d" % next_task_id)
        fake_task._decorator_files_re(*unnamed_args, **named_args)
        fake_task._complete_setup()
        return fake_task.param_generator_func, fake_task

    def files_re(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        return list(p1 for (p1, ps) in self.get_param_iterator(*unnamed_args, **named_args)[0](None))
        # return list(self.get_param_iterator (*old_args)(None))

    def check_input_files_exist(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        it = self.get_param_iterator(*unnamed_args, **named_args)[0]
        for param, param2 in it(None):
            check_input_files_exist(*param)
        return True

    def needs_update_check_modify_time(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        it, task = self.get_param_iterator(*unnamed_args, **named_args)
        #print >> sys.stderr, [p for (p, param2) in it(None)], "??"
        return [needs_update_check_modify_time(*p, task=task,
                                               job_history=open_job_history(history_file)) for (p, param2) in it(None)]

    def test_combine(self):
        """
        test combining operator
        """
        paths = self.files_re(tempdir + "/*", r"(.*).test$",
                              combine(r"\1.input"), r"\1.output")
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [
            (('DIR/f0.input',), 'DIR/f0.output'),
                             (('DIR/f1.input',), 'DIR/f1.output'),
                             (('DIR/f2.input',), 'DIR/f2.output'),
        ]
        )
        paths = self.files_re(tempdir + "/*", "(.*).test$",
                              combine(r"\1.input"), r"combined.output")
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [(('DIR/f0.input',
                            'DIR/f1.input',
                            'DIR/f2.input'), 'combined.output')])

    def test_glob(self):
        """
        test globbed form
        """
        #
        # simple 1 input, 1 output
        #
        paths = self.files_re(tempdir + "/*", "(.*).test$",
                              r"\1.input", r"\1.output")
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [('DIR/f0.input', 'DIR/f0.output'),
                          ('DIR/f1.input', 'DIR/f1.output'),
                          ('DIR/f2.input', 'DIR/f2.output')])
        self.assertTrue(self.check_input_files_exist(tempdir + "/*", "(.*).test$",
                                                     r"\1.test", r"\1.output"))

        #
        # nested forms
        #
        paths = self.files_re(tempdir + "/*", "(.*).test$",
                              [r"\1.input", 2, ["something", r"\1"]], r"\1.output", r"\1.extra", 5)
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [(['DIR/f0.input', 2, ['something', 'DIR/f0']], 'DIR/f0.output', 'DIR/f0.extra', 5),
                          (['DIR/f1.input', 2, ['something', 'DIR/f1']],
                           'DIR/f1.output', 'DIR/f1.extra', 5),
                          (['DIR/f2.input', 2, ['something', 'DIR/f2']], 'DIR/f2.output', 'DIR/f2.extra', 5)])

        #
        # only output
        #
        paths = self.files_re(tempdir + "/*", ".*/(.*).test$", r"\1.output")
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [('DIR/f0.test', 'f0.output'),
                             ('DIR/f1.test', 'f1.output'),
                             ('DIR/f2.test', 'f2.output')])

    def test_globbed_up_to_date(self):
        """
        test glob form
        """
        #
        # check simple is up to date
        #
        self.assertEqual([res[0] for res in self.needs_update_check_modify_time(tempdir + "/*",
                                                                                "(.*).test$", r"\1.output")], [True, False, False])
        #
        # check complex is up to date
        #
        self.assertEqual([res[0] for res in self.needs_update_check_modify_time(tempdir + "/*",
                                                                                "(.*).test$", [1, 2, [[r"\1.output",
                                                                                                       r"\1.output"]]])], [True, False, False])

    def test_filelist(self):
        """
        test file list form
        """
        file_list = ["DIR/f0.test", "DIR/f1.test", "DIR/f2.test"]
        #
        # simple 1 input, 1 output
        #
        paths = self.files_re(file_list, r"(.*).test$",
                              r"\1.input", r"\1.output")
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [('DIR/f0.input', 'DIR/f0.output'),
                          ('DIR/f1.input', 'DIR/f1.output'),
                          ('DIR/f2.input', 'DIR/f2.output')])

        #
        # nested forms
        #
        paths = self.files_re(
            file_list, "(.*).test$", [r"\1.input", 2, ["something", r"\1"]], r"\1.output", r"\1.extra", 5)
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [(['DIR/f0.input', 2, ['something', 'DIR/f0']], 'DIR/f0.output', 'DIR/f0.extra', 5),
                          (['DIR/f1.input', 2, ['something', 'DIR/f1']],
                           'DIR/f1.output', 'DIR/f1.extra', 5),
                          (['DIR/f2.input', 2, ['something', 'DIR/f2']], 'DIR/f2.output', 'DIR/f2.extra', 5)])

        #
        # only output
        #
        paths = self.files_re(file_list, ".*/(.*).test$", r"\1.output")
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [('DIR/f0.test', 'f0.output'),
                             ('DIR/f1.test', 'f1.output'),
                             ('DIR/f2.test', 'f2.output')])

    def test_tasks(self):
        """
        test if can use tasks to specify dependencies
        """

        paths = self.files_re(output_from("module.func1",
                                               "module.func2",
                                               "module.func3",
                                               "module.func4",
                                               "module.func5"), r"(.test)", r"\1.yes")
        self.assertEqual(paths,
                         [
                             ((2, 'output5.test'), 'output5.test.yes'),
                             (['output4.test', 'output.ignored'],
                              'output4.test.yes'),
                             ('output1.test', 'output1.test.yes'),
                             ('output2.test', 'output2.test.yes'),
                             ('output3.test', 'output3.test.yes'),
                         ])

        paths = self.files_re(output_from(
            "module.func2"), r"(.ignored)", r"\1.yes")
        self.assertEqual(paths,
                         [('output.ignored', 'output.ignored.yes')])
        paths = self.files_re(
            [output_from("module.func2")], r"(.ignored)", r"\1.yes")
        self.assertEqual(paths,
                         [('output.ignored', 'output.ignored.yes')])


# =========================================================================================

#   split_param_factory

# =========================================================================================

class Test_split_param_factory(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(tempdir):
            os.makedirs(tempdir)
        touch("%s/f%d.output" % (tempdir, 0))
        for i in range(3):
            touch("%s/f%d.test" % (tempdir, i))
        time.sleep(0.1)
        touch("%s/f%d.output" % (tempdir, 1))
        touch("%s/f%d.output" % (tempdir, 2))
        for ii in range(3):
            touch("%s/%d.test_match1" % (tempdir, ii))
            touch("%s/%d.test_match2" % (tempdir, ii))

        self.tasks = [t1, t2, t3, t4, t5]

    def tearDown(self):
        for i in range(3):
            os.unlink("%s/f%d.test" % (tempdir, i))
            os.unlink("%s/%d.test_match1" % (tempdir, i))
            os.unlink("%s/%d.test_match2" % (tempdir, i))
        for i in range(3):
            os.unlink("%s/f%d.output" % (tempdir, i))
        os.removedirs(tempdir)
        pass

    # _____________________________________________________________________________

    #   wrappers

    # _____________________________________________________________________________

    def get_param_iterator(self, *unnamed_args, **named_args):
        #
        # replace function / function names with tasks
        #
        # fake virgin task
        fake_task = task.Task(list_generator_factory,
                              "module.func_fake%d" % randint(1, 1000000))
        fake_task._decorator_split(*unnamed_args, **named_args)
        fake_task._complete_setup()
        return fake_task.param_generator_func

    def do_task_split(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        # extra dereference because we are only interested in the first (only) job
        # return list(self.get_param_iterator (*unnamed_args, **named_args)(None))[0]
        return list(p1 for (p1, ps) in self.get_param_iterator(*unnamed_args, **named_args)(None))[0]

    def test_glob(self):
        """
        test globbed form
        """
        #
        # simple 1 input, 1 output
        #
        paths = self.do_task_split(
            tempdir + "/*", [tempdir + "/*.test_match1", tempdir + "/*.test_match2"])
        self.assertEqual(recursive_replace(recursive_replace(paths, tempdir, "DIR"), exe_path, "DIR"),
                         (['DIR/0.test_match1',
                           'DIR/0.test_match2',
                             'DIR/1.test_match1',
                             'DIR/1.test_match2',
                             'DIR/2.test_match1',
                             'DIR/2.test_match2',
                             'DIR/f0.output',
                             'DIR/f0.test',
                             'DIR/f1.output',
                             'DIR/f1.test',
                             'DIR/f2.output',
                             'DIR/f2.test'],
                          ['DIR/0.test_match1',
                             'DIR/1.test_match1',
                             'DIR/2.test_match1',
                             'DIR/0.test_match2',
                             'DIR/1.test_match2',
                             'DIR/2.test_match2']))

    def test_tasks(self):
        """
        test if can use tasks to specify dependencies
        """

        paths = self.do_task_split([output_from("module.func1",       # input params
                                                     "module.func2",
                                                     "module.func3",
                                                     "module.func4",
                                                     "module.func5"),
                                    tempdir + "/*"],
                                   [tempdir + "/*.test_match1",                    # output params
                                    tempdir + "/*.test_match2",
                                    "extra.file"],
                                   6)                                      # extra params
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR")
        self.assertEqual(paths,
                         ([5,
                           ['output4.test', 'output.ignored'],
                           'output1.test',
                           'output2.test',
                           'output3.test',
                           'output.ignored',
                           (2, 'output5.test'),
                           'DIR/0.test_match1',
                           'DIR/0.test_match2',
                           'DIR/1.test_match1',
                           'DIR/1.test_match2',
                           'DIR/2.test_match1',
                           'DIR/2.test_match2',
                           'DIR/f0.output',
                           'DIR/f0.test',
                           'DIR/f1.output',
                           'DIR/f1.test',
                           'DIR/f2.output',
                           'DIR/f2.test'],
                          ['DIR/0.test_match1',
                             'DIR/1.test_match1',
                             'DIR/2.test_match1',
                             'DIR/0.test_match2',
                             'DIR/1.test_match2',
                             'DIR/2.test_match2',
                             'extra.file'],
                          6))

        # single job output consisting of a single file
        paths = self.do_task_split(output_from(
            "module.func2"), tempdir + "/*.test_match1")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths,  ('output.ignored', [
                         'DIR/0.test_match1', 'DIR/1.test_match1', 'DIR/2.test_match1']))

        paths = self.do_task_split(
            [output_from("module.func2")], tempdir + "/*.test_match1")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths,  (['output.ignored'], [
                         'DIR/0.test_match1', 'DIR/1.test_match1', 'DIR/2.test_match1']))

        # single job output consisting of a list
        paths = self.do_task_split(output_from(
            "module.func4"), tempdir + "/*.test_match1")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths,  ((2, 'output5.test'), [
                         'DIR/0.test_match1', 'DIR/1.test_match1', 'DIR/2.test_match1']))

        paths = self.do_task_split(
            [output_from("module.func4")], tempdir + "/*.test_match1")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, ([(2, 'output5.test')], [
                         'DIR/0.test_match1', 'DIR/1.test_match1', 'DIR/2.test_match1']))

# =========================================================================================

#   merge_param_factory

# =========================================================================================


class Test_merge_param_factory(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(tempdir):
            os.makedirs(tempdir)
        touch("%s/f%d.output" % (tempdir, 0))
        for i in range(3):
            touch("%s/f%d.test" % (tempdir, i))
        time.sleep(0.1)
        touch("%s/f%d.output" % (tempdir, 1))
        touch("%s/f%d.output" % (tempdir, 2))

        self.tasks = [t1, t2, t3, t4, t5]

    def tearDown(self):
        for i in range(3):
            os.unlink("%s/f%d.test" % (tempdir, i))
        for i in range(3):
            os.unlink("%s/f%d.output" % (tempdir, i))
        os.removedirs(tempdir)
        pass

    # _____________________________________________________________________________

    #   wrappers

    # _____________________________________________________________________________

    def get_param_iterator(self, *unnamed_args, **named_args):
        #
        # replace function / function names with tasks
        #
        # fake virgin task
        fake_task = task.Task(list_generator_factory,
                              "module.func_fake%d" % randint(1, 1000000))
        fake_task._decorator_merge(*unnamed_args, **named_args)
        fake_task._complete_setup()
        return fake_task.param_generator_func

    def do_task_merge(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        # extra dereference because we are only interested in the first (only) job
        return list(p1 for (p1, ps) in self.get_param_iterator(*unnamed_args, **named_args)(None))[0]

    def test_glob(self):
        """
        test globbed form
        """
        expected_result = (['DIR/f0.output',
                            'DIR/f0.test',
                            'DIR/f1.output',
                            'DIR/f1.test',
                            'DIR/f2.output',
                            'DIR/f2.test',
                            ],
                           ["test1",
                            "test2",
                            "extra.file"]
                           )
        #
        # simple 1 input, 1 output
        #
        paths = self.do_task_merge(tempdir + "/*",
                                   ["test1",                               # output params
                                    "test2",
                                    "extra.file"])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_result)

        #
        #   named parameters
        #
        paths = self.do_task_merge(input=tempdir + "/*",
                                   output=["test1",                               # output params
                                           "test2",
                                           "extra.file"])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_result)

    def test_tasks(self):
        """
        test if can use tasks to specify dependencies
        """

        unnamed_args = [[output_from("module.func1",       # input params
                                          "module.func2",
                                          "module.func3",
                                          "module.func4",
                                          "module.func5"),
                         tempdir + "/*"],
                        ["test1",                               # output params
                         "test2",
                         "extra.file"],
                        6]                                      # extra params

        expected_results = ([
                            5,
                            ['output4.test', 'output.ignored'],
                            'output1.test',
                            'output2.test',
                            'output3.test',
                            'output.ignored',
                            (2, 'output5.test'),
                            'DIR/f0.output',
                            'DIR/f0.test',
                            'DIR/f1.output',
                            'DIR/f1.test',
                            'DIR/f2.output',
                            'DIR/f2.test'],
                            ['test1',                          # output params
                             'test2',
                             'extra.file'],
                            6)

        # unnamed arguments
        paths = self.do_task_merge(*unnamed_args)
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        # NAMED ARGUMENTS
        paths = self.do_task_merge(input=unnamed_args[0],
                                   output=unnamed_args[1],
                                   extras=unnamed_args[2:])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        paths = self.do_task_merge(output_from(
            "module.func2"), "output", "extra")
        paths = self.do_task_merge(output_from(
            "module.func1", "module.func2"), "output", "extra")

        # single job output consisting of a single file
        paths = self.do_task_merge(output_from("module.func2"), "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, ('output.ignored', 'output'))

        paths = self.do_task_merge(
            [output_from("module.func2")], "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, (['output.ignored'], 'output'))

        # single job output consisting of a list
        paths = self.do_task_merge(output_from("module.func4"), "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, ((2, 'output5.test'), 'output'))

        paths = self.do_task_merge(
            [output_from("module.func4")], "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, ([(2, 'output5.test')], 'output'))

# =========================================================================================

#   transform_param_factory

# =========================================================================================


class Test_transform_param_factory(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(tempdir):
            os.makedirs(tempdir)
        touch("%s/f%d.output" % (tempdir, 0))
        for i in range(3):
            touch("%s/f%d.test" % (tempdir, i))
        time.sleep(0.1)
        touch("%s/f%d.output" % (tempdir, 1))
        touch("%s/f%d.output" % (tempdir, 2))

        self.tasks = [t1, t2, t3, t4, t5]
        self.maxDiff = None

    def tearDown(self):
        for i in range(3):
            os.unlink("%s/f%d.test" % (tempdir, i))
        for i in range(3):
            os.unlink("%s/f%d.output" % (tempdir, i))
        os.removedirs(tempdir)
        pass

    # _____________________________________________________________________________

    #   wrappers

    # _____________________________________________________________________________

    def get_param_iterator(self, *unnamed_args, **named_args):
        #
        # replace function / function names with tasks
        #
        # fake virgin task
        fake_task = task.Task(list_generator_factory,
                              "module.func_fake%d" % randint(1, 1000000))
        fake_task._decorator_transform(*unnamed_args, **named_args)
        fake_task._complete_setup()
        return fake_task.param_generator_func

    def do_task_transform(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        # extra dereference because we are only interested in the first (only) job
        return list(p1 for (p1, ps) in self.get_param_iterator(*unnamed_args, **named_args)(None))

    def test_simple(self):
        """
        test simple_form
        """
        #
        # simple 1 input, 1 output
        #
        paths = self.do_task_transform("a.test", task.regex("a(.+)"),  r"b\1")

        self.assertEqual(paths,
                         [('a.test', 'b.test')])

    def test_suffix(self):
        """
        test suffix transform with globs
        """
        #
        # simple 1 input, 1 output
        #
        paths = self.do_task_transform(tempdir + "/*.test", task.suffix(".test"),
                                       [".output1", ".output2"], ".output3")

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             ('DIR/f0.test', ['DIR/f0.output1',
                                              'DIR/f0.output2'], ".output3"),
                             ('DIR/f1.test', ['DIR/f1.output1',
                                              'DIR/f1.output2'], ".output3"),
                             ('DIR/f2.test', ['DIR/f2.output1',
                                              'DIR/f2.output2'], ".output3"),
                         ])

    def test_formatter(self):
        """
        test suffix transform with globs
        """
        #
        # simple 1 input, 1 output
        #
        unnamed_args = [tempdir + "/*.test",
                        task.formatter("/(?P<name>\w+).test$"),
                        ["{path[0]}/{name[0]}.output1{ext[0]}", "{path[0]}/{name[0]}.output2"], "{path[0]}/{name[0]}.output3"]
        expected_results = [
            ('DIR/f0.test', ['DIR/f0.output1.test',
                             'DIR/f0.output2'], "DIR/f0.output3"),
            ('DIR/f1.test', ['DIR/f1.output1.test',
                             'DIR/f1.output2'], "DIR/f1.output3"),
            ('DIR/f2.test', ['DIR/f2.output1.test', 'DIR/f2.output2'], "DIR/f2.output3"), ]
        # unnamed_args                                   ]
        paths = self.do_task_transform(*unnamed_args)
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        # named args
        paths = self.do_task_transform(
            input=unnamed_args[0], filter=unnamed_args[1], output=unnamed_args[2], extras=unnamed_args[3:])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

    def test_regex(self):
        """
        test regex transform with globs
        """
        #
        # simple 1 input, 1 output
        #
        unnamed_args = [tempdir + "/*.test",
                        task.regex(r"(.*)\.test"),
                        [r"\1.output1", r"\1.output2"], r"\1.output3"]
        expected_results = [
            ('DIR/f0.test', ['DIR/f0.output1',
                             'DIR/f0.output2'], "DIR/f0.output3"),
            ('DIR/f1.test', ['DIR/f1.output1',
                             'DIR/f1.output2'], "DIR/f1.output3"),
            ('DIR/f2.test', ['DIR/f2.output1', 'DIR/f2.output2'], "DIR/f2.output3"), ]

        # unnamed_args                                   ]
        paths = self.do_task_transform(*unnamed_args)
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        # named args
        paths = self.do_task_transform(
            input=unnamed_args[0], filter=unnamed_args[1], output=unnamed_args[2], extras=unnamed_args[3:])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

    def test_inputs(self):
        """
        test transform with inputs in both regex and suffix forms
        """
        #
        # simple 1 input, 1 output
        #
        #
        paths = self.do_task_transform(tempdir + "/*.test", task.regex(r"(.*)\.test"),
                                       task.inputs(r"\1.testwhat"),
                                       [r"\1.output1", r"\1.output2"])

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             ('DIR/f0.testwhat',
                              ['DIR/f0.output1', 'DIR/f0.output2']),
                             ('DIR/f1.testwhat',
                                 ['DIR/f1.output1', 'DIR/f1.output2']),
                             ('DIR/f2.testwhat',
                                 ['DIR/f2.output1', 'DIR/f2.output2']),
                         ])
        paths = self.do_task_transform(tempdir + "/*.test", task.suffix(".test"),
                                       task.inputs(r"a.testwhat"),
                                       [".output1", ".output2"], ".output3")

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             ('a.testwhat', ['DIR/f0.output1',
                                             'DIR/f0.output2'], '.output3'),
                             ('a.testwhat', ['DIR/f1.output1',
                                             'DIR/f1.output2'], '.output3'),
                             ('a.testwhat', ['DIR/f2.output1', 'DIR/f2.output2'], '.output3')])
        #
        # add inputs
        #
        #
        unnamed_args = [tempdir + "/*.test", task.regex(r"(.*)\.test"),
                        add_inputs(r"\1.testwhat"),
                        [r"\1.output1", r"\1.output2"]]
        expected_results = [
            (('DIR/f0.test', 'DIR/f0.testwhat'),
             ['DIR/f0.output1', 'DIR/f0.output2']),
            (('DIR/f1.test', 'DIR/f1.testwhat'),
             ['DIR/f1.output1', 'DIR/f1.output2']),
            (('DIR/f2.test', 'DIR/f2.testwhat'),
             ['DIR/f2.output1', 'DIR/f2.output2']),
        ]

        # unnamed_args                                   ]
        paths = self.do_task_transform(*unnamed_args)
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        # named args
        paths = self.do_task_transform(
            input=unnamed_args[0], filter=unnamed_args[1], add_inputs=unnamed_args[2], output=unnamed_args[3], extras=unnamed_args[4:])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        # named args
        paths = self.do_task_transform(
            *unnamed_args[0:2], add_inputs=unnamed_args[2].args, output=unnamed_args[3], extras=unnamed_args[4:])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        paths = self.do_task_transform(tempdir + "/*.test", task.suffix(".test"),
                                       add_inputs(r"a.testwhat"),
                                       [".output1", ".output2"], ".output3")

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             (('DIR/f0.test', 'a.testwhat'),
                              ['DIR/f0.output1', 'DIR/f0.output2'], '.output3'),
                             (('DIR/f1.test', 'a.testwhat'),
                                 ['DIR/f1.output1', 'DIR/f1.output2'], '.output3'),
                             (('DIR/f2.test', 'a.testwhat'), ['DIR/f2.output1', 'DIR/f2.output2'], '.output3')])

    def test_tasks(self):
        """
        test if can use tasks to specify dependencies
        """

        paths = self.do_task_transform([output_from("module.func1",       # input params
                                                         "module.func2",
                                                         "module.func3",
                                                         "module.func4",
                                                         "module.func5"),
                                        tempdir + "/*.test"],
                                       task.regex(r"(.*)\.test"),
                                       [r"\1.output1", r"\1.output2"], r"\1.output3")
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             ((2, 'output5.test'), [
                              'output5.output1', 'output5.output2'], 'output5.output3'),
                             ('DIR/f0.test', ['DIR/f0.output1',
                                              'DIR/f0.output2'], 'DIR/f0.output3'),
                             ('DIR/f1.test', ['DIR/f1.output1',
                                              'DIR/f1.output2'], 'DIR/f1.output3'),
                             ('DIR/f2.test', ['DIR/f2.output1',
                                              'DIR/f2.output2'], 'DIR/f2.output3'),
                             (['output4.test', 'output.ignored'], [
                                 'output4.output1', 'output4.output2'], 'output4.output3'),
                             ('output1.test', [
                                 'output1.output1', 'output1.output2'], 'output1.output3'),
                             ('output2.test', [
                                 'output2.output1', 'output2.output2'], 'output2.output3'),
                             ('output3.test', [
                                 'output3.output1', 'output3.output2'], 'output3.output3'),
                         ])

        # single job output consisting of a single file
        paths = self.do_task_transform(output_from(
            "module.func2"), task.regex(r"(.*)\..*"),  r"\1.output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, [('output.ignored', 'output.output')])

        # Same output if task specified as part of a list of tasks
        paths = self.do_task_transform(
            [output_from("module.func2")], task.regex(r"(.*)\..*"),  "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, [('output.ignored', 'output')])

        # single job output consisting of a list
        paths = self.do_task_transform(output_from(
            "module.func4"), task.regex(r"(.*)\..*"),  "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, [((2, 'output5.test'), 'output')])

        # Same output if task specified as part of a list of tasks
        paths = self.do_task_transform(
            [output_from("module.func4")], task.regex(r"(.*)\..*"),  "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, [((2, 'output5.test'), 'output')])

#
#
# =========================================================================================

#   collate_param_factory

# =========================================================================================


class Test_collate_param_factory(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(tempdir):
            os.makedirs(tempdir)
        touch("%s/f%d.output" % (tempdir, 0))
        touch("%s/e%d.output" % (tempdir, 0))
        for i in range(3):
            touch("%s/f%d.test" % (tempdir, i))
        for i in range(3):
            touch("%s/e%d.test" % (tempdir, i))
        time.sleep(0.1)
        touch("%s/f%d.output" % (tempdir, 1))
        touch("%s/f%d.output" % (tempdir, 2))
        touch("%s/e%d.output" % (tempdir, 1))
        touch("%s/e%d.output" % (tempdir, 2))

        self.tasks = [t1, t2, t3, t4, t5]

    def tearDown(self):
        for i in range(3):
            os.unlink("%s/f%d.test" % (tempdir, i))
            os.unlink("%s/e%d.test" % (tempdir, i))
        for i in range(3):
            os.unlink("%s/f%d.output" % (tempdir, i))
            os.unlink("%s/e%d.output" % (tempdir, i))
        os.removedirs(tempdir)
        pass

    # _____________________________________________________________________________

    #   wrappers

    # _____________________________________________________________________________

    def get_param_iterator(self, *unnamed_args, **named_args):
        #
        # replace function / function names with tasks
        #
        # fake virgin task
        fake_task = task.Task(list_generator_factory,
                              "module.func_fake%d" % randint(1, 1000000))
        fake_task._decorator_collate(*unnamed_args, **named_args)
        fake_task._complete_setup()
        return fake_task.param_generator_func

    def do_task_collate(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        # extra dereference because we are only interested in the first (only) job
        return list(p1 for (p1, ps) in self.get_param_iterator(*unnamed_args, **named_args)(None))

    def test_regex(self):
        """
        test regex collate with globs
        """
        paths = self.do_task_collate(
            tempdir + "/*", task.regex(r"(.*).test$"), r"\1.output")
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [
            (('DIR/e0.test',), 'DIR/e0.output'),
            (('DIR/e1.test',), 'DIR/e1.output'),
            (('DIR/e2.test',), 'DIR/e2.output'),
            (('DIR/f0.test',), 'DIR/f0.output'),
            (('DIR/f1.test',), 'DIR/f1.output'),
            (('DIR/f2.test',), 'DIR/f2.output'),
        ]
        )
        paths = self.do_task_collate(
            tempdir + "/*", task.regex("(.*).test$"), task.inputs(r"\1.input2"), r"combined.output")
        self.assertEqual(recursive_replace(paths, tempdir, "DIR"),
                         [((
                             'DIR/e0.input2',
                             'DIR/e1.input2',
                             'DIR/e2.input2',
                             'DIR/f0.input2',
                             'DIR/f1.input2',
                             'DIR/f2.input2',
                         ), 'combined.output')])

        #
        # simple 1 input, 1 output
        #
        paths = self.do_task_collate(tempdir + "/*.test", task.regex(r"(.*/[ef]).*\.test"),
                                     [r"\1.output1", r"\1.output2"], r"\1.extra")

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             (
                                 ('DIR/e0.test', 'DIR/e1.test',
                                  'DIR/e2.test'),      # input
                                 # output
                                 ['DIR/e.output1', 'DIR/e.output2'],
                                 'DIR/e.extra'                                       # extra
                             ),
                             (
                                 ('DIR/f0.test', 'DIR/f1.test',
                                  'DIR/f2.test'),      # input
                                 # output
                                 ['DIR/f.output1', 'DIR/f.output2'],
                                 'DIR/f.extra'                                       # extra
                             )
                         ])

    def test_inputs(self):
        """
        test collate with task.inputs
        """
        #
        # collating using inputs
        #
        paths = self.do_task_collate(tempdir + "/*.test", task.regex(r"(.*/[ef])(.).*\.test"),
                                     task.inputs(r"\1\2.whoopee"),  [r"\1.output1", r"\1.output2"], r"\1.extra")

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             (
                                 ('DIR/e0.whoopee', 'DIR/e1.whoopee',
                                  'DIR/e2.whoopee'),      # input
                                 # output
                                 ['DIR/e.output1', 'DIR/e.output2'],
                                 'DIR/e.extra'                                                # extra
                             ),
                             (
                                 ('DIR/f0.whoopee', 'DIR/f1.whoopee',
                                  'DIR/f2.whoopee'),      # input
                                 # output
                                 ['DIR/f.output1', 'DIR/f.output2'],
                                 'DIR/f.extra'                                                # extra
                             )
                         ])
        #
        # collating using inputs where some files do not match regex
        #
        paths = self.do_task_collate(tempdir + "/*.test", task.regex(r"(.*/f)[a-z0-9]+\.test"),
                                     task.inputs(r"\1.whoopee"),  [r"\1.output1", r"\1.output2"], r"\1.extra")

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [(('DIR/f.whoopee',), ['DIR/f.output1', 'DIR/f.output2'], 'DIR/f.extra')])

        #
        # collating using inputs where multiple copies of the same input names are removed
        #
        paths = self.do_task_collate(tempdir + "/*.test", task.regex(r"(.*/[ef])[a-z0-9]+\.test"),
                                     task.inputs(r"\1.whoopee"),  [r"\1.output1", r"\1.output2"], r"\1.extra")

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             (
                                 # input
                                 ('DIR/e.whoopee',),
                                 # output
                                 ['DIR/e.output1', 'DIR/e.output2'],
                                 'DIR/e.extra'                                      # extra
                             ),
                             (
                                 # input
                                 ('DIR/f.whoopee',),
                                 # output
                                 ['DIR/f.output1', 'DIR/f.output2'],
                                 'DIR/f.extra'                                      # extra
                             )
                         ])

        #
        #   test python set object. Note that set is constructed with the results of the substitution
        #

        unnamed_args = [tempdir + "/*.test",
                        task.regex(r"(.*/[ef])[a-z0-9]+\.test"),
                        task.inputs(r"\1.whoopee"),
                        set([r"\1.output1", r"\1.output2", tempdir + "/e.output2"]),
                        r"\1.extra"]
        expected_results = [
            (
                # input
                ('DIR/e.whoopee',),
                # output
                set(['DIR/e.output1', 'DIR/e.output2']),
                'DIR/e.extra'                                      # extra
            ),
            (
                # input
                ('DIR/f.whoopee',),
                set(['DIR/f.output1', 'DIR/f.output2',
                     'DIR/e.output2']),                # output
                'DIR/f.extra'                                      # extra
            )
        ]

        # unnamed_args                                   ]
        paths = self.do_task_collate(*unnamed_args)
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        # named args
        paths = self.do_task_collate(input=unnamed_args[0], filter=unnamed_args[1],
                                     replace_inputs=unnamed_args[2], output=unnamed_args[3], extras=unnamed_args[4:])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        # named args
        paths = self.do_task_collate(
            *unnamed_args[0:2], replace_inputs=unnamed_args[2], output=unnamed_args[3], extras=unnamed_args[4:])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

        # named args
        paths = self.do_task_collate(
            *unnamed_args[0:2], replace_inputs=unnamed_args[2].args[0], output=unnamed_args[3], extras=unnamed_args[4:])
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths, expected_results)

    def test_tasks(self):
        """
        test if can use tasks to specify dependencies
        """

        paths = self.do_task_collate([output_from("module.func1",       # input params
                                                       "module.func2",
                                                       "module.func3",
                                                       "module.func4",
                                                       "module.func5"),
                                      tempdir + "/*.test"],
                                     task.regex(r"(.*[oef])[a-z0-9]+\.test"),
                                     [r"\1.output1", r"\1.output2"], r"\1.extra")
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             (('DIR/e0.test', 'DIR/e1.test', 'DIR/e2.test'),
                              ['DIR/e.output1', 'DIR/e.output2'], 'DIR/e.extra'),
                             (('DIR/f0.test', 'DIR/f1.test', 'DIR/f2.test'),
                                 ['DIR/f.output1', 'DIR/f.output2'], 'DIR/f.extra'),
                             (((2, 'output5.test'), ['output4.test', 'output.ignored'], 'output1.test', 'output2.test', 'output3.test'), [
                                 'o.output1', 'o.output2'], 'o.extra')
                         ])

        paths = self.do_task_collate([output_from("module.func1",       # input params
                                                       "module.func2",
                                                       "module.func3",
                                                       "module.func4",
                                                       "module.func5"),
                                      tempdir + "/*.test"],
                                     task.regex(r"(.*[oef])[a-z0-9]+\.test"),
                                     task.inputs(r"\1.whoopee"),
                                     [r"\1.output1", r"\1.output2"], r"\1.extra")
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             (
                                 # input
                                 ("DIR/e.whoopee",),
                                 # output
                                 ["DIR/e.output1", "DIR/e.output2"],
                                 "DIR/e.extra"                                   # extra
                             ),
                             (
                                 # input
                                 ("DIR/f.whoopee",),
                                 # output
                                 ["DIR/f.output1", "DIR/f.output2"],
                                 "DIR/f.extra"                                   # extra
                             ),
                             (
                                 # input
                                 ("o.whoopee",),
                                 # output
                                 ["o.output1", "o.output2"],
                                 "o.extra"                                       # extra
                             )
                         ])

        # single job output consisting of a single file
        paths = self.do_task_collate(output_from(
            "module.func2"), task.regex(r"(.*)\..*"),  r"\1.output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        # print dumps(paths, indent = 4)
        self.assertEqual(paths, [(('output.ignored',), 'output.output')])

        # Same output if task specified as part of a list of tasks
        paths = self.do_task_collate(
            [output_from("module.func2")], task.regex(r"(.*)\..*"),  "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, [(('output.ignored',), 'output')])

        #
        # single job output consisting of a list
        #
        paths = self.do_task_collate(output_from(
            "module.func4"), task.regex(r"(.*)\..*"),  "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, [(((2, 'output5.test'),), 'output')])

        # Same output if task specified as part of a list of tasks
        paths = self.do_task_collate(
            [output_from("module.func4")], task.regex(r"(.*)\..*"),  "output")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths, [(((2, 'output5.test'),), 'output')])

# =========================================================================================

#   files_param_factory

# =========================================================================================


class Test_files_param_factory(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(tempdir):
            os.makedirs(tempdir)
        touch("%s/f%d.output" % (tempdir, 0))
        for i in range(3):
            touch("%s/f%d.test" % (tempdir, i))
        time.sleep(0.1)
        touch("%s/f%d.output" % (tempdir, 1))
        touch("%s/f%d.output" % (tempdir, 2))

        self.tasks = [t1, t2, t3, t4, t5]

    def tearDown(self):
        for i in range(3):
            os.unlink("%s/f%d.test" % (tempdir, i))
        for i in range(3):
            os.unlink("%s/f%d.output" % (tempdir, i))
        os.removedirs(tempdir)
        pass

    def get_param_iterator(self, *unnamed_args, **named_args):
        #
        # replace function / function names with tasks
        #
        # fake virgin task
        fake_task = task.Task(list_generator_factory,
                              "module.func_fake%d" % randint(1, 1000000))
        fake_task._decorator_files(*unnamed_args, **named_args)
        fake_task._complete_setup()
        return fake_task.param_generator_func

    def files(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        return list(p1 for (p1, ps) in self.get_param_iterator(*unnamed_args, **named_args)(None))

    def _test_simple(self):
        """
        test simple_form
        """
        #
        # simple 1 input, 1 output
        #
        paths = self.files("a.test", "b.test")
        self.assertEqual(paths,
                         [('a.test', 'b.test')])

    def test_glob(self):
        """
        test globbed form
        """
        #
        # Replacement of globs in first parameter
        #
        paths = self.files(tempdir + "/*", "a.test", "b.test")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths,
                         [
                             (
                                 [
                                     "DIR/f0.output",
                                     "DIR/f0.test",
                                     "DIR/f1.output",
                                     "DIR/f1.test",
                                     "DIR/f2.output",
                                     "DIR/f2.test"
                                 ],
                                 "a.test",
                                 "b.test"
                             )
                         ])
        #
        # Replacement of globs in first parameter in-place
        #
        paths = self.files([tempdir + "/*", "robbie.test"], "a.test", "b.test")
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths,
                         [
                             (
                                 [
                                     "DIR/f0.output",
                                     "DIR/f0.test",
                                     "DIR/f1.output",
                                     "DIR/f1.test",
                                     "DIR/f2.output",
                                     "DIR/f2.test",
                                     "robbie.test"
                                 ],
                                 "a.test",
                                 "b.test"
                             )
                         ])
        #
        # No Replacement of globs in other parameter of multi-job task
        #
        paths = self.files([[[tempdir + "/*", "robbie.test"],
                             "a.test", "b.test"], ["a.test", ["b.test", 2], "a.*"]])
        paths = recursive_replace(recursive_replace(
            paths, tempdir, "DIR"), exe_path, "DIR_E")
        self.assertEqual(paths,
                         [
                             (
                                 [
                                     "DIR/f0.output",
                                     "DIR/f0.test",
                                     "DIR/f1.output",
                                     "DIR/f1.test",
                                     "DIR/f2.output",
                                     "DIR/f2.test",
                                     "robbie.test"
                                 ],
                                 "a.test",
                                 "b.test"
                             ),
                             ("a.test", ["b.test", 2], "a.*")
                         ])

    def _test_filelist(self):
        """
        test file list form
        """
        # simple list
        file_list = ["DIR/f0.test", "DIR/f1.test", "DIR/f2.test"]
        paths = self.files(*file_list)
        self.assertEqual(paths, [
            ("DIR/f0.test", "DIR/f1.test", "DIR/f2.test"),
        ])

        # complex list
        file_list = [[["DIR/e0.test", set([5, "DIR/e1.test"]), "DIR/e2.test"], [
            "DIR/f0.test", "DIR/f1.test", "DIR/f2.test"]]]
        paths = self.files(*file_list)
        self.assertEqual(paths, [
            ("DIR/e0.test", set([5, "DIR/e1.test"]), "DIR/e2.test"),
            ("DIR/f0.test", "DIR/f1.test", "DIR/f2.test"),
        ])

        # bad list: missing list enclosure
        file_list = [["DIR/f0.test", "DIR/f1.test", "DIR/f2.test"]]
        self.assertRaises(error_task_files, self.files, *file_list)

        # bad list: missing io
        file_list = [[[1, 2], ["DIR/e0.test", [5, "DIR/e1.test"],
                               "DIR/e2.test"], ["DIR/f0.test", "DIR/f1.test", "DIR/f2.test"]]]
        self.assertRaises(error_task_files, self.files, *file_list)

    def _test_tasks(self):
        """
        test if can use tasks to specify dependencies
        """
        #
        #   substitution of tasks
        #
        paths = self.files(output_from("module.func1",
                                            "module.func2",
                                            "module.func3",
                                            "module.func4",
                                            "module.func5"), "rob.test", "b.test")
        self.assertEqual(paths, [([5, ['output4.test', 'output.ignored'], 'output1.test', 'output2.test', 'output3.test',
                                   'output.ignored', (2, 'output5.test')],
                                  'rob.test',
                                  'b.test')])

        #
        #   nested in place substitution of tasks
        #
        paths = self.files([output_from("module.func1",
                                             "module.func2",
                                             "module.func3",
                                             "module.func4",
                                             "module.func5"),
                            "robbie.test"], "a.test", "b.test")
        self.assertEqual(paths,
                         [([5, ['output4.test', 'output.ignored'], 'output1.test', 'output2.test', 'output3.test', 'output.ignored', (2, 'output5.test'), 'robbie.test'], 'a.test', 'b.test')])

        # single job output consisting of a single file
        paths = self.files(output_from("module.func2"), "output", "extra")
        self.assertEqual(paths, [('output.ignored', 'output', 'extra')])

        # Different output if task specified as part of a list of tasks
        paths = self.files([output_from("module.func2"),
                            output_from("module.func2")], "output", "extra")
        self.assertEqual(
            paths, [(['output.ignored', 'output.ignored'], 'output', 'extra')])

        # single job output consisting of a list
        paths = self.files(output_from("module.func4"), "output", "extra")
        self.assertEqual(paths, [((2, 'output5.test'), 'output', 'extra')])

        # Same output if task specified as part of a list of tasks
        paths = self.files([output_from("module.func4"),
                            output_from("module.func2")], "output", "extra")
        self.assertEqual(
            paths,  [([(2, 'output5.test'), 'output.ignored'], 'output', 'extra')])


#
# =========================================================================================

#   product_param_factory

# =========================================================================================

class Test_product_param_factory(unittest.TestCase):
    def setUp(self):
        if not os.path.exists(tempdir):
            os.makedirs(tempdir)
        touch("%s/a.test1" % (tempdir))
        touch("%s/b.test1" % (tempdir))
        touch("%s/c.test2" % (tempdir))
        touch("%s/d.test2" % (tempdir))
        touch("%s/a.testwhat1" % (tempdir))
        touch("%s/b.testwhat1" % (tempdir))
        touch("%s/c.testwhat2" % (tempdir))
        touch("%s/d.testwhat2" % (tempdir))
        time.sleep(0.1)
        touch("%s/a.b.output" % (tempdir))
        touch("%s/a.c.output" % (tempdir))
        touch("%s/b.c.output" % (tempdir))
        touch("%s/b.d.output" % (tempdir))

        self.tasks = [t1, t2, t3, t4, t5]
        self.maxDiff = None

    def tearDown(self):
        os.unlink("%s/a.test1" % (tempdir))
        os.unlink("%s/b.test1" % (tempdir))
        os.unlink("%s/c.test2" % (tempdir))
        os.unlink("%s/d.test2" % (tempdir))
        os.unlink("%s/a.testwhat1" % (tempdir))
        os.unlink("%s/b.testwhat1" % (tempdir))
        os.unlink("%s/c.testwhat2" % (tempdir))
        os.unlink("%s/d.testwhat2" % (tempdir))
        os.unlink("%s/a.b.output" % (tempdir))
        os.unlink("%s/a.c.output" % (tempdir))
        os.unlink("%s/b.c.output" % (tempdir))
        os.unlink("%s/b.d.output" % (tempdir))

        os.removedirs(tempdir)
        pass

    # _____________________________________________________________________________

    #   wrappers

    # _____________________________________________________________________________

    def get_param_iterator(self, *unnamed_args, **named_args):
        #
        # replace function / function names with tasks
        #
        # fake virgin task
        fake_task = task.Task(list_generator_factory,
                              "module.func_fake%d" % randint(1, 1000000))
        fake_task._decorator_product(*unnamed_args, **named_args)
        fake_task._complete_setup()
        return fake_task.param_generator_func

    def do_task_product(self, *unnamed_args, **named_args):
        """
        This extra function is to simulate the forwarding from the decorator to
            the task creation function
        """
        # extra dereference because we are only interested in the first (only) job
        return list(p1 for (p1, ps) in self.get_param_iterator(*unnamed_args, **named_args)(None))

    def test_simple(self):
        """
        test simple_form
        """
        #
        # simple 1 input, 1 output
        #
        args = [[tempdir + "/a.test1", tempdir + "/b.test1"],
                task.formatter("(?:.+/)?(?P<ID>\w+)\.(.+)"),
                [tempdir + "/c.test2", tempdir + "/d.test2", tempdir + "/e.ignore"],
                task.formatter("(?:.+/)?(?P<ID>\w+)\.(test2)"),
                r"{path[0][0]}/{ID[0][0]}.{1[1][0]}.output"]
        expected_result = [
            (('DIR/a.test1', 'DIR/c.test2'), 'DIR/a.c.output'),
            (('DIR/a.test1', 'DIR/d.test2'), 'DIR/a.d.output'),
            (('DIR/b.test1', 'DIR/c.test2'), 'DIR/b.c.output'),
            (('DIR/b.test1', 'DIR/d.test2'), 'DIR/b.d.output')
        ]
        paths = self.do_task_product(*args)
        self.assertEqual(recursive_replace(
            paths, tempdir, "DIR"), expected_result)

        # named parameters
        paths = self.do_task_product(input=args[0], filter=args[1],
                                     input2=args[2], filter2=args[3],
                                     output=args[4])
        self.assertEqual(recursive_replace(
            paths, tempdir, "DIR"), expected_result)

        # named parameters
        paths = self.do_task_product(*args[0:2],
                                     input2=args[2], filter2=args[3],
                                     output=args[4])
        self.assertEqual(recursive_replace(
            paths, tempdir, "DIR"), expected_result)

        paths = self.do_task_product(*args[0:4],
                                     output=args[4])
        self.assertEqual(recursive_replace(
            paths, tempdir, "DIR"), expected_result)

    def test_inputs(self):
        """
        test transform with inputs in both regex and suffix forms
        """
        #
        # (replace) inputs
        #
        #
        paths = self.do_task_product([tempdir + "/a.test1", tempdir + "/b.test1"],                          task.formatter("(?:.+/)?(?P<ID>\w+)\.(.+)"),
                                     [tempdir + "/c.test2", tempdir + "/d.test2", tempdir +
                                         "/e.ignore"], task.formatter("(?:.+/)?(?P<ID>\w+)\.(test2)"),
                                     task.inputs(
                                         ("{path[0][0]}/{basename[0][0]}.testwhat1", "{path[1][0]}/{basename[1][0]}.testwhat2")),
                                     r"{path[0][0]}/{ID[0][0]}.{1[1][0]}.output")
        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             (('DIR/a.testwhat1', 'DIR/c.testwhat2'), 'DIR/a.c.output'),
                             (('DIR/a.testwhat1', 'DIR/d.testwhat2'), 'DIR/a.d.output'),
                             (('DIR/b.testwhat1', 'DIR/c.testwhat2'), 'DIR/b.c.output'),
                             (('DIR/b.testwhat1', 'DIR/d.testwhat2'), 'DIR/b.d.output')
                         ]
                         )
        #
        # add inputs
        #
        #
        paths = self.do_task_product([tempdir + "/a.test1", tempdir + "/b.test1"],                          task.formatter("(?:.+/)?(?P<ID>\w+)\.(.+)"),
                                     [tempdir + "/c.test2", tempdir + "/d.test2", tempdir +
                                         "/e.ignore"], task.formatter("(?:.+/)?(?P<ID>\w+)\.(test2)"),
                                     add_inputs(
                                         "{path[0][0]}/{basename[0][0]}.testwhat1", "{path[1][0]}/{basename[1][0]}.testwhat2", ),
                                     r"{path[0][0]}/{ID[0][0]}.{1[1][0]}.output")

        paths = recursive_replace(paths, tempdir, "DIR")
        self.assertEqual(paths,
                         [
                             ((('DIR/a.test1', 'DIR/c.test2'), 'DIR/a.testwhat1',
                               'DIR/c.testwhat2'), 'DIR/a.c.output'),
                             ((('DIR/a.test1', 'DIR/d.test2'), 'DIR/a.testwhat1',
                               'DIR/d.testwhat2'), 'DIR/a.d.output'),
                             ((('DIR/b.test1', 'DIR/c.test2'), 'DIR/b.testwhat1',
                               'DIR/c.testwhat2'), 'DIR/b.c.output'),
                             ((('DIR/b.test1', 'DIR/d.test2'), 'DIR/b.testwhat1',
                               'DIR/d.testwhat2'), 'DIR/b.d.output')
                         ]
                         )


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import json
import unittest
from ruffus import pipeline_run, pipeline_printout, Pipeline, originate, transform, suffix, \
    follows, runtime_parameter, mkdir
import ruffus
import shutil
import sys

"""

    test_runtime_data.py

        Tests having inputs depend on a run time parameter
        (named element in the dict runtime_data passed to pipeline_run)

"""


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"
runtime_files = [tempdir + "a.3"]


# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# use simplejson in place of json for python < 2.6
# try:
#    import json
# except ImportError:
#    import simplejson
#    json = simplejson


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#
#    task1
#
@originate([tempdir + 'a.1'] + runtime_files)
def task1(outfile):
    """
    First task
    """
    output_text = ""
    output_text += "    -> " + json.dumps(outfile) + "\n"
    with open(outfile, "w") as oo:
        oo.write(output_text)


#
#    task2
#
@transform(task1, suffix(".1"), ".2")
def task2(infile, outfile):
    """
    Second task
    """
    if infile:
        with open(infile) as ii:
            output_text = ii.read()
    else:
        output_text = ""
    output_text += json.dumps(infile) + " -> " + json.dumps(outfile) + "\n"
    with open(outfile, "w") as oo:
        oo.write(output_text)


#
#    task3
#
@transform(task2, suffix(".2"), ".3")
def task3(infile, outfile):
    """
    Third task
    """
    if infile:
        with open(infile) as ii:
            output_text = ii.read()
    else:
        output_text = ""
    output_text += json.dumps(infile) + " -> " + json.dumps(outfile) + "\n"
    with open(outfile, "w") as oo:
        oo.write(output_text)


#
#    task4
#
@follows(task3)
@transform(runtime_parameter("a"), suffix(".3"), ".4")
def task4(infile, outfile):
    """
    Fourth task
    """
    if infile:
        with open(infile) as ii:
            output_text = ii.read()
    else:
        output_text = ""
    output_text += json.dumps(infile) + " -> " + json.dumps(outfile) + "\n"
    with open(outfile, "w") as oo:
        oo.write(output_text)


class Test_ruffus(unittest.TestCase):
    def setUp(self):
        for f in ["a.1", "a.2", "a.3", "a.4"]:
            f = os.path.join(tempdir, f)
            if os.path.exists(f):
                os.unlink(f)
        try:
            os.makedirs(tempdir)
        except:
            pass

    def tearDown(self):
        for f in ["a.1", "a.2", "a.3", "a.4"]:
            f = os.path.join(tempdir, f)
            if os.path.exists(f):
                os.unlink(f)
            else:
                raise Exception("%s is missing" % f)

        shutil.rmtree(tempdir)

    def test_ruffus(self):
        pipeline_run(verbose=0, runtime_data={
                     "a": runtime_files}, pipeline="main")

    def test_newstyle_ruffus(self):

        test_pipeline = Pipeline("test")
        test_pipeline.originate(task_func=task1,
                                output=[tempdir + 'a.1'] + runtime_files)
        test_pipeline.transform(task2, task1, suffix(".1"), ".2")
        test_pipeline.transform(task_func=task3,
                                input=task2,
                                filter=suffix(".2"),
                                output=".3")
        test_pipeline.transform(task_func=task4,
                                input=runtime_parameter("a"),
                                filter=suffix(".3"),
                                output=".4").follows(task3)
        test_pipeline.run(verbose=0, runtime_data={"a": runtime_files})


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import unittest
from ruffus import originate, transform, Pipeline, pipeline_run, suffix
import sys

"""
    test_softlink_uptodate.py
"""


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#
#   First task
#
@originate(["a.1", "b.1"])
def start_task(output_file_name):
    with open(output_file_name,  "w") as f:
        pass

#
#   Forwards file names, is always as up to date as its input files...
#


@transform(start_task, suffix(".1"), ".1")
def same_file_name_task(input_file_name, output_file_name):
    pass

#
#   Links file names, is always as up to date if links are not missing
#


@transform(start_task, suffix(".1"), ".linked.1")
def linked_file_name_task(input_file_name, output_file_name):
    try:
        os.symlink(input_file_name, output_file_name)
    except:
        print(input_file_name, output_file_name)
        raise


#
#   Final task linking everything
#
@transform([linked_file_name_task, same_file_name_task], suffix(".1"), ".3")
def final_task(input_file_name, output_file_name):
    with open(output_file_name,  "w") as f:
        pass

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Run pipeline


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


try:
    from StringIO import StringIO
except:
    from io import StringIO


class Test_ruffus(unittest.TestCase):
    def setUp(self):
        for f in ["a.1", "b.1", "a.linked.1", "b.linked.1", "a.3", "b.3", "a.linked.3", "b.linked.3"]:
            try:
                if os.path.exists(f):
                    os.unlink(f)
            except:
                print("    !!!!OOPs. Can't unlink %s" % f, file=sys.stderr)
                pass

    def tearDown(self):
        for f in ["a.1", "b.1", "a.linked.1", "b.linked.1", "a.3", "b.3", "a.linked.3", "b.linked.3"]:
            if os.path.lexists(f):
                os.unlink(f)
            else:
                raise Exception("Expected %s missing" % f)

    def test_ruffus(self):
        pipeline_run(log_exceptions=True, verbose=0, pipeline="main")

    def test_newstyle_ruffus(self):
        test_pipeline = Pipeline("test")
        test_pipeline.originate(start_task, ["a.1", "b.1"])
        test_pipeline.transform(same_file_name_task,
                                start_task, suffix(".1"), ".1")
        test_pipeline.transform(linked_file_name_task,
                                start_task, suffix(".1"), ".linked.1")
        test_pipeline.transform(
            final_task, [linked_file_name_task, same_file_name_task], suffix(".1"), ".3")
        test_pipeline.run(log_exceptions=True, verbose=0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
import sys, os

# add self to search path for testing
if __name__ == '__main__':
    exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
    module_name = os.path.split(sys.argv[0])[1]
    module_name = os.path.splitext(module_name)[0];
else:
    module_name = __name__

# Use import path from <<../python_modules>>
if __name__ == '__main__':
    sys.path.insert(0, os.path.abspath(os.path.join(exe_path,"../..")))

from ruffus import *
import json


def Explicitly_specified_task(infile, outfile):
    pass


@follows(Explicitly_specified_task)
def Task_to_run(infile, outfile):
    pass

@follows(Task_to_run)
def Up_to_date_task_forced_to_rerun(infile, outfile):
    pass


@follows(Up_to_date_task_forced_to_rerun)
def Final_target(infile, outfile):
    pass

@follows(Final_target)
def Downstream_task1_ignored(infile, outfile):
    pass

@follows(Final_target)
def Downstream_task2_ignored(infile, outfile):
    pass

pipeline_printout_graph ("pretty_flowchart.png", "png", [Final_target], [Explicitly_specified_task], dpi = 72, size = (2,2), draw_vertically = False, no_key_legend = True)




#!/usr/bin/env python
"""

    test2.py
    [--log_file PATH]
    [--verbose]

"""

################################################################################
#
#   test2
#
#
#   Copyright (c) 7/16/2010 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################

import sys, os

# add self to search path for testing
if __name__ == '__main__':
    exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
    module_name = os.path.split(sys.argv[0])[1]
    module_name = os.path.splitext(module_name)[0];
else:
    module_name = __name__

# Use import path from <<../python_modules>>
if __name__ == '__main__':
    sys.path.insert(0, os.path.abspath(os.path.join(exe_path,"../..")))



#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


if __name__ == '__main__':
    from optparse import OptionParser
    import StringIO

    parser = OptionParser(version="%prog 1.0", usage = "\n\n    %progs [options]")
    parser.add_option("-i", "--input_file", dest="input_file",
                      metavar="FILE",
                      type="string",
                      help="Name and path of input file. "
                          "Defaults to reading from STDIN.")

    #
    #   general options: verbosity / logging
    #
    parser.add_option("-v", "--verbose", dest = "verbose",
                      action="count", default=0,
                      help="Print more verbose messages for each additional verbose level.")
    parser.add_option("-L", "--log_file", dest="log_file",
                      metavar="FILE",
                      type="string",
                      help="Name and path of log file")
    parser.add_option("--skip_parameter_logging", dest="skip_parameter_logging",
                        action="store_true", default=False,
                        help="Do not print program parameters to log.")
    parser.add_option("--debug", dest="debug",
                        action="count", default=0,
                        help="Set default program parameters in debugging mode.")







    #
    #   pipeline
    #
    parser.add_option("-t", "--target_tasks", dest="target_tasks",
                        action="append",
                        default = list(),
                        metavar="JOBNAME",
                        type="string",
                        help="Target task(s) of pipeline.")
    parser.add_option("-j", "--jobs", dest="jobs",
                        default=1,
                        metavar="N",
                        type="int",
                        help="Allow N jobs (commands) to run simultaneously.")
    parser.add_option("-n", "--just_print", dest="just_print",
                        action="store_true", default=False,
                        help="Don't actually run any commands; just print the pipeline.")
    parser.add_option("--flowchart", dest="flowchart",
                        metavar="FILE",
                        type="string",
                        help="Don't actually run any commands; just print the pipeline "
                             "as a flowchart.")

    #
    #   Less common pipeline options
    #
    parser.add_option("--key_legend_in_graph", dest="key_legend_in_graph",
                        action="store_true", default=False,
                        help="Print out legend and key for dependency graph.")
    parser.add_option("--draw_graph_horizontally", dest="draw_horizontally",
                        action="store_true", default=False,
                        help="Draw horizontal dependency graph.")
    parser.add_option("--forced_tasks", dest="forced_tasks",
                        action="append",
                        default = list(),
                        metavar="JOBNAME",
                        type="string",
                        help="Pipeline task(s) which will be included even if they are up to date.")

    # get help string
    f =StringIO.StringIO()
    parser.print_help(f)
    helpstr = f.getvalue()
    original_args = " ".join(sys.argv)
    (options, remaining_args) = parser.parse_args()


    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    #                                             #
    #   Debug: Change these                       #
    #                                             #
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
    if options.debug:
        options.log_file                = os.path.join("test2.log")
        options.verbose                 = 5
        options.log_parameters          = True
    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv
    #                                             #
    #   Debug: Change these                       #
    #                                             #
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    #
    #   mandatory options
    #
    mandatory_options = []
    def check_mandatory_options (options, mandatory_options, helpstr):
        """
        Check if specified mandatory options have b een defined
        """
        missing_options = []
        for o in mandatory_options:
            if not getattr(options, o):
                missing_options.append("--" + o)

        if not len(missing_options):
            return

        raise Exception("Missing mandatory parameter%s: %s.\n\n%s\n\n" %
                        ("s" if len(missing_options) > 1 else "",
                         ", ".join(missing_options),
                         helpstr))
    check_mandatory_options (options, mandatory_options, helpstr)


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

from ruffus import *
from ruffus.ruffus_exceptions import JobSignalledBreak

#from json import dumps
#from collections import defaultdict



#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Logger


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

if __name__ == '__main__':
    import logging
    import logging.handlers

    MESSAGE = 15
    logging.addLevelName(MESSAGE, "MESSAGE")

    def setup_std_logging (logger, log_file, verbose):
        """
        set up logging using programme options
        """
        class debug_filter(logging.Filter):
            """
            Ignore INFO messages
            """
            def filter(self, record):
                return logging.INFO != record.levelno

        class NullHandler(logging.Handler):
            """
            for when there is no logging
            """
            def emit(self, record):
                pass

        # We are interesting in all messages
        logger.setLevel(logging.DEBUG)
        has_handler = False

        # log to file if that is specified
        if log_file:
            handler = logging.FileHandler(log_file, delay=False)
            handler.setFormatter(logging.Formatter("%(asctime)s - %(name)s - %(levelname)6s - %(message)s"))
            handler.setLevel(MESSAGE)
            logger.addHandler(handler)
            has_handler = True

        # log to stderr if verbose
        if verbose:
            stderrhandler = logging.StreamHandler(sys.stderr)
            stderrhandler.setFormatter(logging.Formatter("    %(message)s"))
            stderrhandler.setLevel(logging.DEBUG)
            if log_file:
                stderrhandler.addFilter(debug_filter())
            logger.addHandler(stderrhandler)
            has_handler = True

        # no logging
        if not has_handler:
            logger.addHandler(NullHandler())


    #
    #   set up log
    #
    logger = logging.getLogger(module_name)
    setup_std_logging(logger, options.log_file, options.verbose)

    #
    #   Allow logging across Ruffus pipeline
    #
    def get_logger (logger_name, args):
        return logger

    from ruffus.proxy_logger import *
    (logger_proxy,
     logging_mutex) = make_shared_logger_and_proxy (get_logger,
                                                    module_name,
                                                    {})

    #
    #   log programme parameters
    #
    if not options.skip_parameter_logging:
        programme_name = os.path.split(sys.argv[0])[1]
        logger.info("%s %s" % (programme_name, original_args))


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Pipeline


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
@follows("task2")
@files(None, "a.1")
def task1(input_file, output_file):
    open(output_file, "w")

@transform(task1,  suffix("1"),  "2")
def task2(input_file, output_file):
    open(output_file, "w")

@transform(task2,  suffix("2"),  "3")
def task3(input_file, output_file):
    open(output_file, "w")

@transform(task3,  suffix("3"),  "4")
def task4(input_file, output_file):
    open(output_file, "w")

@transform(task4,  suffix("4"),  "5")
def task5(input_file, output_file):
    open(output_file, "w")

@transform(task5,  suffix("5"),  "6")
def task6(input_file, output_file):
    open(output_file, "w")

@transform(task1,  suffix("1"),  "7")
def task7(input_file, output_file):
    open(output_file, "w")

@transform(task7,  suffix("7"),  "8")
def task8(input_file, output_file):
    open(output_file, "w")

@transform(task8,  suffix("8"),  "9")
def task9(input_file, output_file):
    open(output_file, "w")

@follows("task5")
@follows("task12")
@transform(task9,  suffix("9"),  "10")
def task10(input_file, output_file):
    open(output_file, "w")

@transform(task10, suffix("10"), "11")
def task11(input_file, output_file):
    open(output_file, "w")

@follows(task6)
@transform(task11, suffix("11"), "12")
def task12(input_file, output_file):
    open(output_file, "w")

@transform(task12, suffix("12"), "13")
def task13(input_file, output_file):
    open(output_file, "w")

@transform(task13, suffix("13"), "14")
def task14(input_file, output_file):
    open(output_file, "w")

@follows(task9)
@transform(task14, suffix("14"), "15")
def task15(input_file, output_file):
    open(output_file, "w")

@follows(task11)
@transform(task15, suffix("15"), "16")
def task16(input_file, output_file):
    open(output_file, "w")

@transform(task16, suffix("16"), "17")
def task17(input_file, output_file):
    open(output_file, "w")

@transform(task12, suffix("12"), "18")
def task18(input_file, output_file):
    open(output_file, "w")

@transform(task18, suffix("18"), "19")
def task19(input_file, output_file):
    open(output_file, "w")

@transform(task19, suffix("19"), "20")
def task20(input_file, output_file):
    open(output_file, "w")

@follows(task3)
@transform(task20, suffix("20"), "21")
def task21(input_file, output_file):
    open(output_file, "w")

@transform(task21, suffix("21"), "22")
def task22(input_file, output_file):
    open(output_file, "w")

@transform(task22, suffix("22"), "23")
def task23(input_file, output_file):
    open(output_file, "w")

@transform(task23, suffix("23"), "24")
def task24(input_file, output_file):
    open(output_file, "w")

@follows(task17)
@transform(task24, suffix("24"), "25")
def task25(input_file, output_file):
    open(output_file, "w")
#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
if __name__ == '__main__':
#   debug code not run if called as a module
#
    if options.just_print:
        pipeline_printout(sys.stdout, options.target_tasks, options.forced_tasks,
                            verbose=options.verbose)

    elif options.flowchart:
        pipeline_printout_graph (     open(options.flowchart, "w"),
                             os.path.splitext(options.flowchart)[1][1:],
                             options.target_tasks,
                             options.forced_tasks,
                             draw_vertically = not options.draw_horizontally,
                             no_key_legend   = not options.key_legend_in_graph,
                              user_colour_scheme = {"colour_scheme_index":0},
                              pipeline_name = "Complicated pipeline:",
                              size          = (6,5),
                              dpi           = 72,
                                      )
    else:
        pipeline_run(options.target_tasks, options.forced_tasks,
                            multiprocess    = options.jobs,
                            logger          = stderr_logger,
                            verbose         = options.verbose)











#!/usr/bin/env python
from __future__ import print_function
import re
import shutil
import unittest
from ruffus.ruffus_utility import CHECKSUM_FILE_TIMESTAMPS, RUFFUS_HISTORY_FILE, get_default_history_file_name
from ruffus.ruffus_exceptions import RethrownJobError
from ruffus import pipeline_run, pipeline_printout, Pipeline, collate, mkdir, regex, \
    suffix, formatter, originate, transform
import sys
import os

"""

    test_job_history_with_exceptions.py

        Make sure that when an exception is thrown only the current and following tasks fail

"""

# sub-1s resolution in system?
one_second_per_job = None
throw_exception = False

tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#   imports
# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

try:
    from StringIO import StringIO
except:
    from io import StringIO


# ___________________________________________________________________________
#
#   generate_initial_files1
# ___________________________________________________________________________
@originate([tempdir + prefix + "_name.tmp1" for prefix in "abcd"])
def generate_initial_files1(on):
    with open(on, 'w') as outfile:
        pass

# ___________________________________________________________________________
#
#   generate_initial_files2
# ___________________________________________________________________________


@originate([tempdir + "e_name.tmp1", tempdir + "f_name.tmp1"])
def generate_initial_files2(on):
    with open(on, 'w') as outfile:
        pass

# ___________________________________________________________________________
#
#   generate_initial_files3
# ___________________________________________________________________________


@originate([tempdir + "/g_name.tmp1", tempdir + "/h_name.tmp1"])
def generate_initial_files3(on):
    with open(on, 'w') as outfile:
        pass

# ___________________________________________________________________________
#
#   generate_initial_files1
# ___________________________________________________________________________


@originate(tempdir + "i_name.tmp1")
def generate_initial_files4(on):
    with open(on, 'w') as outfile:
        pass

# ___________________________________________________________________________
#
#   check_task2
# ___________________________________________________________________________


@collate([generate_initial_files1, generate_initial_files2, generate_initial_files3,
          generate_initial_files4],
         formatter(),
         "{path[0]}/all.tmp2")
# @transform([generate_initial_files1, generate_initial_files2, generate_initial_files3,
#            generate_initial_files4],
#            formatter( ),
#            "{path[0]}/{basename[0]}.tmp2")
def check_task2(infiles, outfile):
    with open(outfile, "w") as p:
        pass
    #print >>sys.stderr, "8" * 80, "\n", "    task2 :%s %s " % (infiles, outfile)

# ___________________________________________________________________________
#
#   check_task3
# ___________________________________________________________________________


@transform(check_task2, suffix(".tmp2"), ".tmp3")
def check_task3(infile, outfile):
    global throw_exception
    if throw_exception != None:
        throw_exception = not throw_exception
    if throw_exception:
        #print >>sys.stderr, "Throw exception for ", infile, outfile
        raise Exception("oops")
    else:
        #print >>sys.stderr, "No throw exception for ", infile, outfile
        pass
    with open(outfile, "w") as p:
        pass
    #print >>sys.stderr, "8" * 80, "\n", "    task3 :%s %s " % (infile, outfile)

# ___________________________________________________________________________
#
#   check_task4
# ___________________________________________________________________________


@transform(check_task3, suffix(".tmp3"), ".tmp4")
def check_task4(infile, outfile):
    with open(outfile, "w") as p:
        pass
    #print >>sys.stderr, "8" * 80, "\n", "    task4 :%s %s " % (infile, outfile)


def cleanup_tmpdir():
    os.system('rm -f %s %s' %
              (os.path.join(tempdir, '*'), RUFFUS_HISTORY_FILE))


VERBOSITY = 5
VERBOSITY = 11

cnt_pipelines = 0


class Test_job_history_with_exceptions(unittest.TestCase):
    def setUp(self):
        try:
            os.mkdir(tempdir)
        except OSError:
            pass

    # ___________________________________________________________________________
    #
    #   test product() pipeline_printout and pipeline_run
    # ___________________________________________________________________________
    def test_job_history_with_exceptions(self):
        cleanup_tmpdir()
        s = StringIO()
        pipeline_printout(s, [check_task4], verbose=VERBOSITY,
                          wrap_width=10000, pipeline="main")
        # print s.getvalue()

    def create_pipeline(self):
        # each pipeline has a different name
        global cnt_pipelines
        cnt_pipelines = cnt_pipelines + 1
        test_pipeline = Pipeline("test %d" % cnt_pipelines)

        test_pipeline.originate(task_func=generate_initial_files1,
                                output=[tempdir + prefix + "_name.tmp1" for prefix in "abcd"])

        test_pipeline.originate(task_func=generate_initial_files2,
                                output=[tempdir + "e_name.tmp1", tempdir + "f_name.tmp1"])

        test_pipeline.originate(task_func=generate_initial_files3,
                                output=[tempdir + "g_name.tmp1", tempdir + "h_name.tmp1"])

        test_pipeline.originate(task_func=generate_initial_files4,
                                output=tempdir + "i_name.tmp1")

        test_pipeline.collate(task_func=check_task2,
                              input=[generate_initial_files1,
                                     generate_initial_files2,
                                     generate_initial_files3,
                                     generate_initial_files4],
                              filter=formatter(),
                              output="{path[0]}/all.tmp2")

        test_pipeline.transform(task_func=check_task3,
                                input=check_task2,
                                filter=suffix(".tmp2"),
                                output=".tmp3")

        test_pipeline.transform(task_func=check_task4,
                                input=check_task3,
                                filter=suffix(".tmp3"),
                                output=".tmp4")
        return test_pipeline

    def test_job_history_with_exceptions_run(self):
        """Run"""
        for i in range(1):
            cleanup_tmpdir()
            try:
                pipeline_run([check_task4], verbose=0,
                             #multithread = 2,
                             one_second_per_job=one_second_per_job, pipeline="main")
            except:
                pass
            s = StringIO()
            pipeline_printout(
                s, [check_task4], verbose=VERBOSITY, wrap_width=10000, pipeline="main")
            #
            # task 2 should be up to date because exception was throw in task 3
            #
            pipeline_printout_str = s.getvalue()
            correct_order = not re.search(
                'Tasks which will be run:.*\n(.*\n)*Task = check_task2', pipeline_printout_str)
            if not correct_order:
                print(pipeline_printout_str)
            self.assertTrue(correct_order)
            sys.stderr.write(".")
        print()

    def test_newstyle_recreate_job_history(self):
        """Run"""
        test_pipeline = self.create_pipeline()
        global throw_exception
        throw_exception = None
        cleanup_tmpdir()

        #
        #      print "Initial run without creating sqlite file"
        #
        test_pipeline.run([check_task4], verbose=0,
                          checksum_level=CHECKSUM_FILE_TIMESTAMPS,
                          multithread=10,
                          one_second_per_job=one_second_per_job)

        #
        #   print "printout without sqlite"
        #
        s = StringIO()
        test_pipeline.printout(
            s, [check_task4], checksum_level=CHECKSUM_FILE_TIMESTAMPS)
        self.assertTrue(not re.search(
            'Tasks which will be run:.*\n(.*\n)*Task = ', s.getvalue()))
        #
        # print "printout expecting sqlite file"
        #
        s = StringIO()
        test_pipeline.printout(s, [check_task4])
        self.assertTrue(
            re.search('Tasks which will be run:.*\n(.*\n)*Task = ', s.getvalue()))
        #
        #   print "Regenerate sqlite file"
        #
        test_pipeline.run([check_task4],
                          checksum_level=CHECKSUM_FILE_TIMESTAMPS,
                          history_file=get_default_history_file_name(),
                          multithread=1,
                          verbose=0,
                          touch_files_only=2,
                          one_second_per_job=one_second_per_job)
        #
        # print "printout expecting sqlite file"
        #
        s = StringIO()
        test_pipeline.printout(s, [check_task4], verbose=VERBOSITY)
        succeed = not re.search(
            'Tasks which will be run:.*\n(.*\n)*Task = ', s.getvalue())
        if not succeed:
            print(s.getvalue(), file=sys.stderr)
        self.assertTrue(succeed)

        throw_exception = False

    #
    def test_newstyle_job_history_with_exceptions_run(self):
        """Run"""
        test_pipeline = self.create_pipeline()
        for i in range(1):
            cleanup_tmpdir()
            try:
                test_pipeline.run([check_task4], verbose=0,
                                  #multithread = 2,
                                  one_second_per_job=one_second_per_job)
            except:
                pass
            s = StringIO()
            test_pipeline.printout(
                s, [check_task4], verbose=VERBOSITY, wrap_width=10000)
            #
            # task 2 should be up to date because exception was throw in task 3
            #
            pipeline_printout_str = s.getvalue()
            correct_order = not re.search(
                'Tasks which will be run:.*\n(.*\n)*Task = check_task2', pipeline_printout_str)
            if not correct_order:
                print(pipeline_printout_str)
            self.assertTrue(correct_order)
            sys.stderr.write(".")
        print()

    def test_recreate_job_history(self):
        """Run"""
        global throw_exception
        throw_exception = None
        cleanup_tmpdir()

        #
        #      print "Initial run without creating sqlite file"
        #
        pipeline_run([check_task4], verbose=0,
                     checksum_level=CHECKSUM_FILE_TIMESTAMPS,
                     multithread=10,
                     one_second_per_job=one_second_per_job, pipeline="main")

        #
        #   print "printout without sqlite"
        #
        s = StringIO()
        pipeline_printout(
            s, [check_task4], checksum_level=CHECKSUM_FILE_TIMESTAMPS, pipeline="main")
        self.assertTrue(not re.search(
            'Tasks which will be run:.*\n(.*\n)*Task = ', s.getvalue()))
        #
        # print "printout expecting sqlite file"
        #
        s = StringIO()
        pipeline_printout(s, [check_task4], pipeline="main")
        self.assertTrue(
            re.search('Tasks which will be run:.*\n(.*\n)*Task = ', s.getvalue()))
        #
        #   print "Regenerate sqlite file"
        #
        pipeline_run([check_task4],
                     checksum_level=CHECKSUM_FILE_TIMESTAMPS,
                     history_file=get_default_history_file_name(),
                     multithread=1,
                     verbose=0,
                     touch_files_only=2,
                     one_second_per_job=one_second_per_job, pipeline="main")
        #
        # print "printout expecting sqlite file"
        #
        s = StringIO()
        pipeline_printout(s, [check_task4], verbose=VERBOSITY, pipeline="main")
        succeed = not re.search(
            'Tasks which will be run:.*\n(.*\n)*Task = ', s.getvalue())
        if not succeed:
            print(s.getvalue(), file=sys.stderr)
        self.assertTrue(succeed)

        throw_exception = False

    # ___________________________________________________________________________
    #
    #   cleanup
    # ___________________________________________________________________________
    def tearDown(self):
        shutil.rmtree(tempdir)
        pass


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    #pipeline_printout(sys.stdout, [check_product_task], verbose = VERBOSITY, pipeline= "main")
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
from ruffus.drmaa_wrapper import write_job_script_to_temp_file, read_stdout_stderr_from_files
import ruffus.drmaa_wrapper
import ruffus
import sys

"""

    test_drmaa_wrapper_run_job_locally.py

"""

import os
script_dir = os.path.abspath(os.path.dirname(__file__))
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0]))

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
class Test_ruffus(unittest.TestCase):

    class t_test_logger:

        """
        Does nothing!
        """

        def __init__(self):
            self.clear()

        def clear(self):
            self.info_msg = []
            self.debug_msg = []
            self.warning_msg = []
            self.error_msg = []

        def info(self, message, *args, **kwargs):
            self.info_msg.append(message)

        def debug(self, message, *args, **kwargs):
            self.debug_msg.append(message)

        def warning(self, message, *args, **kwargs):
            self.warning_msg.append(message)

        def error(self, message, *args, **kwargs):
            self.error_msg.append(message)

    # ___________________________________________________________________________
    #
    #   setup and cleanup
    # ___________________________________________________________________________

    def setUp(self):
        try:
            os.mkdir(tempdir)
        except OSError:
            pass

    def tearDown(self):
        shutil.rmtree(tempdir)

    def test_read_stdout_stderr_from_files(self):
        #
        #   Test missing stdout and stderr files
        #
        stdout_path = os.path.join(tempdir, "stdout.txt")
        stderr_path = os.path.join(tempdir, "stderr.txt")
        logger = Test_ruffus.t_test_logger()
        read_stdout_stderr_from_files(
            stdout_path, stderr_path, logger, cmd_str="test_cmd", tries=0)
        self.assertTrue("could not open stdout" in "".join(logger.warning_msg))
        self.assertTrue("could not open stderr" in "".join(logger.warning_msg))
        logger.clear()

        #
        #   Test present stdout and stderr files
        #
        with open(stdout_path, "w") as so:
            so.write("STDOUT\nSTDOUT\n")
        with open(stderr_path, "w") as se:
            se.write("STDERR\nSTDERR\n")

        stdout_msg, stderr_msg = read_stdout_stderr_from_files(
            stdout_path, stderr_path, logger, cmd_str="test_cmd", tries=1)
        self.assertEqual(logger.warning_msg, [])
        self.assertEqual(stdout_msg, ["STDOUT\n", "STDOUT\n"])
        self.assertEqual(stderr_msg, ["STDERR\n", "STDERR\n"])

    def test_run_job(self):
        environ = {"RUFFUS_HEEHEE": "what?"}
        home_dir = os.path.expanduser("~")
        sys.stderr.write("    Run echoing to screen...\n")
        stdout, stderr = ruffus.drmaa_wrapper.run_job(cmd_str="python %s/slow_process_for_testing.py" % script_dir,
                                                      job_environment=environ,
                                                      working_directory=home_dir,
                                                      run_locally=True,
                                                      verbose=1,
                                                      local_echo=True)
        sys.stderr.write("    Run silently...\n")
        stdout, stderr = ruffus.drmaa_wrapper.run_job(cmd_str="python %s/slow_process_for_testing.py" % script_dir,
                                                      job_environment=environ,
                                                      working_directory=home_dir,
                                                      run_locally=True,
                                                      verbose=1,
                                                      local_echo=False)
        self.assertEqual(
            stdout,
            ['    Stdout 0\n', '    Stdout 1\n', '    Stdout 2\n', '    Stdout 3\n'])

        stderr_fixed = [stderr[x] for x in (0, 2, 3, 4, 5)]
        stderr_variable = stderr[1]
        self.assertEqual(
            stderr_fixed,
            ['     %s\n' % home_dir,
             '    Stderr 0\n',
             '    Stderr 1\n',
             '    Stderr 2\n',
             '    Stderr 3\n'])
        self.assertTrue("'PWD': '{}'".format(home_dir) in stderr_variable)
        self.assertTrue("'RUFFUS_HEEHEE': 'what?'" in stderr_variable)

    def test_write_job_script_to_temp_file(self):
        sys.stderr.write("    Write to temp_file...\n")
        job_script_path, stdout_path, stderr_path = write_job_script_to_temp_file(
            "ls", None, "job_name", "", None, None)
        os.unlink(job_script_path)
        job_script_path, stdout_path, stderr_path = write_job_script_to_temp_file(
            "ls", tempdir, "job_name", "", None, None)

    def test_ls(self):
        sys.stderr.write("    ls...\n")
        with open(os.path.join(tempdir, "temp.txt"), "w") as oo:
            oo.write("done")
        stdout, stderr = ruffus.drmaa_wrapper.run_job(cmd_str="ls %s" % tempdir,
                                                      run_locally=True,
                                                      verbose=1,
                                                      local_echo=True)
        self.assertEqual(stdout, ['temp.txt\n'])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python2.5
"""

    test_tasks.py

"""


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   options


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

from optparse import OptionParser
import sys, os
import os.path
import StringIO

# add self to search path for testing
exe_path = os.path.split(os.path.abspath(sys.argv[0]))[0]
if __name__ == '__main__':
    module_name = os.path.split(sys.argv[0])[1]
    module_name = os.path.splitext(module_name)[0];
else:
    module_name = __name__




parser = OptionParser(version="%prog 1.0")
parser.add_option("-t", "--target_tasks", dest="target_tasks",
                  action="append",
                  default = list(),
                  metavar="JOBNAME",
                  type="string",
                  help="Target task(s) of pipeline.")
parser.add_option("-f", "--forced_tasks", dest="forced_tasks",
                  action="append",
                  default = list(),
                  metavar="JOBNAME",
                  type="string",
                  help="Pipeline task(s) which will be included even if they are up to date.")
parser.add_option("-j", "--jobs", dest="jobs",
                  default=5,
                  metavar="jobs",
                  type="int",
                  help="Specifies  the number of jobs (commands) to run simultaneously.")
parser.add_option("-v", "--verbose", dest = "verbose",
                  action="store_true", default=False,
                  help="Do not echo to shell but only print to log.")
parser.add_option("-d", "--dependency", dest="dependency_file",
                  default="simple.svg",
                  metavar="FILE",
                  type="string",
                  help="Print a dependency graph of the pipeline that would be executed "
                        "to FILE, but do not execute it.")
parser.add_option("-F", "--dependency_graph_format", dest="dependency_graph_format",
                  metavar="FORMAT",
                  type="string",
                  default = 'svg',
                  help="format of dependency graph file. Can be 'ps' (PostScript), "+
                  "'svg' 'svgz' (Structured Vector Graphics), " +
                  "'png' 'gif' (bitmap  graphics) etc ")
parser.add_option("-n", "--just_print", dest="just_print",
                    action="store_true", default=False,
                    help="Print a description of the jobs that would be executed, "
                        "but do not execute them.")
parser.add_option("-M", "--minimal_rebuild_mode", dest="minimal_rebuild_mode",
                    action="store_true", default=False,
                    help="Rebuild a minimum of tasks necessary for the target. "
                    "Ignore upstream out of date tasks if intervening tasks are fine.")
parser.add_option("-K", "--no_key_legend_in_graph", dest="no_key_legend_in_graph",
                    action="store_true", default=False,
                    help="Do not print out legend and key for dependency graph.")
parser.add_option("-H", "--draw_graph_horizontally", dest="draw_horizontally",
                    action="store_true", default=False,
                    help="Draw horizontal dependency graph.")

parameters = [
                ]







#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

import StringIO
import re
import operator
import sys
from collections import defaultdict

sys.path.append(os.path.abspath(os.path.join(exe_path,"..", "..")))
from ruffus import *
import json

#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Functions


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

def create_custom_file_func(params):
    """
    creates function which can be used as input to @files_func
    """
    def cust_func ():
        for job_param in params:
            yield job_param
    return cust_func


def is_job_uptodate (infiles, outfiles, *extra_params):
    """
    assumes first two parameters are files, checks if they are up to date
    """
    return task.needs_update_check_modify_time (infiles, outfiles, *extra_params)



def test_post_task_function ():
    print "Hooray"

import time
def test_job_io(infiles, outfiles, extra_params):
    """
    cat input files content to output files
        after writing out job parameters
    """
    # dump parameters
    params = (infiles, outfiles) + extra_params
    sys.stdout.write('    job = %s\n' % json.dumps(params))



    if isinstance(infiles, str):
        infiles = [infiles]
    elif infiles is None:
        infiles = []
    if isinstance(outfiles, str):
        outfiles = [outfiles]
    output_text = list()
    for f in infiles:
        output_text.append(open(f).read())
    output_text = "".join(sorted(output_text))
    output_text += json.dumps(infiles) + " -> " + json.dumps(outfiles) + "\n"
    for f in outfiles:
        open(f, "w").write(output_text)
    time.sleep(1)


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888





# get help string
f =StringIO.StringIO()
parser.print_help(f)
helpstr = f.getvalue()
(options, remaining_args) = parser.parse_args()





#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


#88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


#
#    task1
#
@files(None, 'a.1')
def task1(infiles, outfiles, *extra_params):
    """
    First task
    """
    test_job_io(infiles, outfiles, extra_params)



#
#    task2
#
@files_re('*.1', '(.*).1', r'\1.1', r'\1.2')
@follows(task1)
def task2(infiles, outfiles, *extra_params):
    """
    Second task
    """
    test_job_io(infiles, outfiles, extra_params)



#
#    task3
#
@files_re('*.1', '(.*).1', r'\1.2', r'\1.3')
@follows(task2)
def task3(infiles, outfiles, *extra_params):
    """
    Third task
    """
    test_job_io(infiles, outfiles, extra_params)



#
#    task4
#
@files_re('*.1', '(.*).1', r'\1.3', r'\1.4')
@follows(task3)
def task4(infiles, outfiles, *extra_params):
    """
    Fourth task
    """
    test_job_io(infiles, outfiles, extra_params)




if options.just_print:
    pipeline_printout(sys.stdout, options.target_tasks, options.forced_tasks,
                        long_winded=True,
                        gnu_make_maximal_rebuild_mode = not options.minimal_rebuild_mode)

elif options.dependency_file:
    pipeline_printout_graph (     open(options.dependency_file, "w"),
                         options.dependency_graph_format,
                         options.target_tasks,
                         options.forced_tasks,
                         draw_vertically = not options.draw_horizontally,
                         gnu_make_maximal_rebuild_mode  = not options.minimal_rebuild_mode,
                         no_key_legend  = options.no_key_legend_in_graph)
else:
    pipeline_run(options.target_tasks, options.forced_tasks, multiprocess = options.jobs,
                    gnu_make_maximal_rebuild_mode  = not options.minimal_rebuild_mode)




#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
import json
from collections import defaultdict
from ruffus.combinatorics import *
from ruffus.ruffus_utility import CHECKSUM_FILE_TIMESTAMPS
from ruffus.ruffus_exceptions import RethrownJobError
from ruffus import pipeline_run, pipeline_printout, Pipeline, follows, merge, posttask, \
    split, collate, mkdir, regex, files_re, combine
import ruffus
import sys

"""

    branching.py

        test branching dependencies

"""

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# use simplejson in place of json for python < 2.6
# try:
#    import json
# except ImportError:
#    import simplejson
#    json = simplejson

# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


species_list = defaultdict(list)
species_list["mammals"].append("cow")
species_list["mammals"].append("horse")
species_list["mammals"].append("sheep")
species_list["reptiles"].append("snake")
species_list["reptiles"].append("lizard")
species_list["reptiles"].append("crocodile")
species_list["fish"].append("pufferfish")


def do_write(file_name, what):
    with open(file_name, "a") as oo:
        oo.write(what)


test_file = tempdir + "task.done"


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    task1
#
@follows(mkdir(tempdir, tempdir + "test"))
@posttask(lambda: do_write(test_file, "Task 1 Done\n"))
def prepare_files():
    for grouping in species_list.keys():
        for species_name in species_list[grouping]:
            filename = tempdir + "%s.%s.animal" % (species_name, grouping)
            with open(filename, "w") as oo:
                oo.write(species_name + "\n")


#
#    task2
#
@files_re(tempdir + '*.animal', r'(.*/)(.*)\.(.*)\.animal', combine(r'\1\2.\3.animal'), r'\1\3.results')
@follows(prepare_files)
@posttask(lambda: do_write(test_file, "Task 2 Done\n"))
def summarise_by_grouping(infiles, outfile):
    """
    Summarise by each species group, e.g. mammals, reptiles, fish
    """
    with open(tempdir + "jobs.start",  "a") as oo:
        oo.write('job = %s\n' % json.dumps([infiles, outfile]))
    with open(outfile, "w") as oo:
        for i in infiles:
            with open(i) as ii:
                oo.write(ii.read())
    with open(tempdir + "jobs.finish",  "a") as oo:
        oo.write('job = %s\n' % json.dumps([infiles, outfile]))


def check_species_correct():
    """
    #cow.mammals.animal
    #horse.mammals.animal
    #sheep.mammals.animal
    #    -> mammals.results
    #
    #snake.reptiles.animal
    #lizard.reptiles.animal
    #crocodile.reptiles.animal
    #    -> reptiles.results
    #
    #pufferfish.fish.animal
    #    -> fish.results
    """
    for grouping in species_list:
        with open(tempdir + grouping + ".results") as ii:
            assert(ii.read() ==
                   "".join(s + "\n" for s in sorted(species_list[grouping])))


class Test_ruffus(unittest.TestCase):

    def tearDown(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    def test_ruffus(self):
        ""
        pipeline_run(multiprocess=10, verbose=0, pipeline="main")
        check_species_correct()


if __name__ == '__main__':
    unittest.main()



################################################################################
#
#   proxy_logger.py
#
#
#   Copyright (c) 10/9/2009 Leo Goodstadt
#
#   Permission is hereby granted, free of charge, to any person obtaining a copy
#   of this software and associated documentation files (the "Software"), to deal
#   in the Software without restriction, including without limitation the rights
#   to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
#   copies of the Software, and to permit persons to whom the Software is
#   furnished to do so, subject to the following conditions:
#
#   The above copyright notice and this permission notice shall be included in
#   all copies or substantial portions of the Software.
#
#   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
#   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
#   AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#   LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
#   OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
#   THE SOFTWARE.
#################################################################################
"""
****************************************************************************
Create proxy for logging for use with multiprocessing
****************************************************************************

These can be safely sent (marshalled) across process boundaries


===========
Example 1
===========

    Set up logger from config file::

        from proxy_logger import *
        args={}
        args["config_file"] = "/my/config/file"

        (logger_proxy,
         logging_mutex) = make_shared_logger_and_proxy (setup_std_shared_logger,
                                                        "my_logger", args)


===========
Example 2
===========

    Log to file ``"/my/lg.log"`` in the specified format (Time / Log name / Event type / Message).

    Delay file creation until first log.

    Only log ``Debug`` messages

        Other alternatives for the logging threshold (``args["level"]``) include

            * ``logging.DEBUG``
            * ``logging.INFO``
            * ``logging.WARNING``
            * ``logging.ERROR``
            * ``logging.CRITICAL``

    ::

        from proxy_logger import *
        args={}
        args["file_name"] = "/my/lg.log"
        args["formatter"] = "%(asctime)s - %(name)s - %(levelname)6s - %(message)s"
        args["delay"]     = True
        args["level"]     = logging.DEBUG

        (logger_proxy,
         logging_mutex) = make_shared_logger_and_proxy (setup_std_shared_logger,
                                                        "my_logger", args)

===========
Example 3
===========

    Rotate log files every 20 Kb, with up to 10 backups.
    ::

        from proxy_logger import *
        args={}
        args["file_name"] = "/my/lg.log"
        args["rotating"] = True
        args["maxBytes"]=20000
        args["backupCount"]=10
        (logger_proxy,
         logging_mutex) = make_shared_logger_and_proxy (setup_std_shared_logger,
                                                        "my_logger", args)



==============
To use:
==============

    ::

        (logger_proxy,
         logging_mutex) = make_shared_logger_and_proxy (setup_std_shared_logger,
                                                        "my_logger", args)

        with logging_mutex:
            my_log.debug('This is a debug message')
            my_log.info('This is an info message')
            my_log.warning('This is a warning message')
            my_log.error('This is an error message')
            my_log.critical('This is a critical error message')
            my_log.log(logging.DEBUG, 'This is a debug message')

    Note that the logging function ``exception()`` is not included because python
    stack trace information is not well-marshalled
    (`pickle <http://docs.python.org/library/pickle.html>`_\ d) across processes.

"""


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

import sys
import os


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Shared logging


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

import multiprocessing
import multiprocessing.managers


import logging
import logging.handlers


#
#   setup_logger
#
def setup_std_shared_logger(logger_name, args):
    """
    This function is a simple around wrapper around the python
    `logging <http://docs.python.org/library/logging.html>`_ module.

    This *logger_factory* example creates logging objects which can
    then be managed by proxy via ``ruffus.proxy_logger.make_shared_logger_and_proxy()``

    This can be:

        * a `disk log file <http://docs.python.org/library/logging.html#filehandler>`_
        * a automatically backed-up `(rotating) log <http://docs.python.org/library/logging.html#rotatingfilehandler>`_.
        * any log specified in a `configuration file <http://docs.python.org/library/logging.html#configuration-file-format>`_

    These are specified in the ``args`` dictionary forwarded by ``make_shared_logger_and_proxy()``

    :param logger_name: name of log
    :param args: a dictionary of parameters forwarded from ``make_shared_logger_and_proxy()``

        Valid entries include:

            .. describe:: "level"

                Sets the `threshold <http://docs.python.org/library/logging.html#logging.Handler.setLevel>`_ for the logger.

            .. describe:: "config_file"

                The logging object is configured from this `configuration file <http://docs.python.org/library/logging.html#configuration-file-format>`_.

            .. describe:: "file_name"

                Sets disk log file name.

            .. describe:: "rotating"

                Chooses a `(rotating) log <http://docs.python.org/library/logging.html#rotatingfilehandler>`_.

            .. describe:: "maxBytes"

                Allows the file to rollover at a predetermined size

            .. describe:: "backupCount"

                If backupCount is non-zero, the system will save old log files by appending the extensions ``.1``, ``.2``, ``.3`` etc., to the filename.

            .. describe:: "delay"

                Defer file creation until the log is written to.

            .. describe:: "formatter"

                `Converts <http://docs.python.org/library/logging.html#formatter-objects>`_ the message to a logged entry string.
                For example,
                ::

                    "%(asctime)s - %(name)s - %(levelname)6s - %(message)s"



    """

    #
    #   Log file name with logger level
    #
    new_logger = logging.getLogger(logger_name)
    if "level" in args:
        new_logger.setLevel(args["level"])

    if "config_file" in args:
        logging.config.fileConfig(args["config_file"])

    else:
        if "file_name" not in args:
            raise Exception(
                "Missing file name for log. Remember to set 'file_name'")
        log_file_name = args["file_name"]

        if "rotating" in args:
            rotating_args = {}
            # override default
            rotating_args["maxBytes"] = args.get("maxBytes", 100000)
            rotating_args["backupCount"] = args.get("backupCount", 5)
            handler = logging.handlers.RotatingFileHandler(
                log_file_name, **rotating_args)
        else:
            defer_loggin = "delay" in args
            handler = logging.handlers.RotatingFileHandler(
                log_file_name, delay=defer_loggin)

        #       %(name)s
        #       %(levelno)s
        #       %(levelname)s
        #       %(pathname)s
        #       %(filename)s
        #       %(module)s
        #       %(funcName)s
        #       %(lineno)d
        #       %(created)f
        #       %(relativeCreated)d
        #       %(asctime)s
        #       %(msecs)d
        #       %(thread)d
        #       %(threadName)s
        #       %(process)d
        #       %(message)s
        #
        # E.g.: "%(asctime)s - %(name)s - %(levelname)6s - %(message)s"
        #
        if "formatter" in args:
            my_formatter = logging.Formatter(args["formatter"])
            handler.setFormatter(my_formatter)

        new_logger.addHandler(handler)

    #
    #   This log object will be wrapped in proxy
    #
    return new_logger


#
#   Proxy object for logging
#       Logging messages will be marshalled (forwarded) to the process where the
#       shared log lives
#
class LoggerProxy(multiprocessing.managers.BaseProxy):
    def debug(self, *args, **kwargs):
        return self._callmethod('debug', args, kwargs)

    def log(self, *args, **kwargs):
        return self._callmethod('log', args, kwargs)

    def info(self, *args, **kwargs):
        return self._callmethod('info', args, kwargs)

    def warning(self, *args, **kwargs):
        return self._callmethod('warning', args, kwargs)

    def error(self, *args, **kwargs):
        return self._callmethod('error', args, kwargs)

    def critical(self, *args, **kwargs):
        return self._callmethod('critical', args, kwargs)

    def log(self, *args, **kwargs):
        return self._callmethod('log', args, kwargs)

    def __str__(self):
        return "<LoggingProxy>"

    def __repr__(self):
        return 'LoggerProxy()'

#
#   Register the setup_logger function as a proxy for setup_logger
#
#   We use SyncManager as a base class so we can get a lock proxy for synchronising
#       logging later on
#


class LoggingManager(multiprocessing.managers.SyncManager):
    """
    Logging manager sets up its own process and will create the real Log object there
    We refer to this (real) log via proxies
    """
    pass


def make_shared_logger_and_proxy(logger_factory, logger_name, args):
    """
    Make a `logging <http://docs.python.org/library/logging.html>`_ object
    called "\ ``logger_name``\ " by calling ``logger_factory``\ (``args``\ )

    This function will return a proxy to the shared logger which can be copied to jobs
    in other processes, as well as a mutex which can be used to prevent simultaneous logging
    from happening.

    :param logger_factory: functions which creates and returns an object with the
                           `logging <http://docs.python.org/library/logging.html>`_ interface.
                           ``setup_std_shared_logger()`` is one example of a logger factory.
    :param logger_name: name of log
    :param args: parameters passed (as a single argument) to ``logger_factory``
    :returns: a proxy to the shared logger which can be copied to jobs in other processes
    :returns: a mutex which can be used to prevent simultaneous logging from happening

    """
    #
    #   make shared log and proxy
    #
    manager = LoggingManager()
    manager.register('setup_logger',
                     logger_factory,
                     proxytype=LoggerProxy,
                     exposed=('critical', 'log',
                              'info', 'debug', 'warning', 'error'))
    manager.start()
    logger_proxy = manager.setup_logger(logger_name, args)

    #
    #   make sure we are not logging at the same time in different processes
    #
    logging_mutex = manager.Lock()

    return logger_proxy, logging_mutex



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
from ruffus import *
import sys

"""

    test_files_post_merge.py

        bug where @files follows merge and extra parenthesis inserted

"""
import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"
test_file = tempdir + "test_output"
jobs_per_task = 50


# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    split_fasta_file
#

def do_write(file_name, what):
    with open(file_name, "a") as oo:
        oo.write(what)


test_file = tempdir + "task.done"


@posttask(lambda: do_write(test_file, "Split into %d files\n" % jobs_per_task))
@split(tempdir + "original.fa", [tempdir + "files.split.success", tempdir + "files.split.*.fa"])
def split_fasta_file(input_file, outputs):

    #
    # remove previous fasta files
    #
    success_flag = outputs[0]
    output_file_names = outputs[1:]
    for f in output_file_names:
        os.unlink(f)

    #
    # create as many files as we are simulating in jobs_per_task
    #
    for i in range(jobs_per_task):
        with open(tempdir + "files.split.%03d.fa" % i, "w") as oo:
            pass

    with open(success_flag,  "w") as oo:
        pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    align_sequences
#
@posttask(lambda: do_write(test_file, "Sequences aligned\n"))
# fa -> aln
@transform(split_fasta_file, suffix(".fa"), ".aln")
def align_sequences(input_file, output_filename):
    with open(output_filename, "w") as oo:
        oo.write("%s\n" % output_filename)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    percentage_identity
#
@posttask(lambda: do_write(test_file, "%Identity calculated\n"))
@transform(align_sequences,             # find all results from align_sequences
           suffix(".aln"),             # replace suffix with:
           [r".pcid",  # .pcid suffix for the result
            r".pcid_success"])  # .pcid_success to indicate job completed
def percentage_identity(input_file, output_files):
    (output_filename, success_flag_filename) = output_files
    with open(output_filename, "w") as oo:
        oo.write("%s\n" % output_filename)
    with open(success_flag_filename, "w") as oo:
        pass


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
#
#    combine_results
#
@posttask(lambda: do_write(test_file, "Results recombined\n"))
@merge(percentage_identity, tempdir + "all.combine_results")
def combine_results(input_files, output_files):
    """
    Combine all
    """
    (output_filename) = output_files
    with open(output_filename, "w") as oo:
        for inp, flag in input_files:
            with open(inp) as ii:
                oo.write(ii.read())


@files(combine_results, os.path.join(tempdir, "check_all_is.well"))
def post_merge_check(input_filename, output_filename):
    """
    check that merge sends just one file, not a list to me
    """
    with open(output_filename, "w") as oo:
        with open(input_filename) as ii:
            oo.write(ii.read())


@files(post_merge_check, os.path.join(tempdir, "check_all_is.weller"))
def post_post_merge_check(input_filename, output_filename):
    """
    check that @files forwards a single file on when given a single file
    """
    with open(output_filename, "w") as oo:
        with open(input_filename) as ii:
            oo.write(ii.read())


try:
    from StringIO import StringIO
except:
    from io import StringIO


class Test_ruffus(unittest.TestCase):
    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        os.makedirs(tempdir)
        open(tempdir + "original.fa", "w").close()
        self.expected_text = """Split into %d files
Sequences aligned
%%Identity calculated
Results recombined
""" % jobs_per_task

    def tearDown(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass

    def test_ruffus(self):
        pipeline_run(multiprocess=50, verbose=0, pipeline="main")
        with open(test_file) as ii:
            post_task_text = ii.read()
        self.assertEqual(post_task_text, self.expected_text)

    def test_newstyle_ruffus(self):

        test_pipeline = Pipeline("test")

        test_pipeline.split(task_func=split_fasta_file,
                            input=tempdir + "original.fa",
                            output=[tempdir + "files.split.success",
                                    tempdir + "files.split.*.fa"])\
            .posttask(lambda: do_write(test_file, "Split into %d files\n" % jobs_per_task))

        test_pipeline.transform(task_func=align_sequences,
                                input=split_fasta_file,
                                filter=suffix(".fa"),
                                output=".aln"                     # fa -> aln
                                )\
            .posttask(lambda: do_write(test_file, "Sequences aligned\n"))

        test_pipeline.transform(task_func=percentage_identity,
                                input=align_sequences,             # find all results from align_sequences
                                # replace suffix with:
                                filter=suffix(".aln"),
                                output=[r".pcid",  # .pcid suffix for the result
                                        r".pcid_success"]  # .pcid_success to indicate job completed
                                )\
            .posttask(lambda: do_write(test_file, "%Identity calculated\n"))

        test_pipeline.merge(task_func=combine_results,
                            input=percentage_identity,
                            output=tempdir + "all.combine_results")\
            .posttask(lambda: do_write(test_file, "Results recombined\n"))

        test_pipeline.files(post_merge_check, combine_results,
                            os.path.join(tempdir, "check_all_is.well"))

        test_pipeline.files(post_post_merge_check, post_merge_check, os.path.join(
            tempdir, "check_all_is.weller"))

        test_pipeline.run(multiprocess=50, verbose=0)
        with open(test_file) as ii:
            post_task_text = ii.read()
        self.assertEqual(post_task_text, self.expected_text)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
from ruffus.ruffus_utility import RUFFUS_HISTORY_FILE, CHECKSUM_FILE_TIMESTAMPS
from ruffus.ruffus_exceptions import RethrownJobError
from ruffus import pipeline_run, pipeline_printout, suffix, transform, split, merge, \
    dbdict, follows, mkdir, originate, Pipeline
import re
import shutil
import sys

"""

    test_verbosity.py

"""

import unittest

import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# funky code to import by file name
parent_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))
ruffus_name = os.path.basename(parent_dir)
ruffus = list(map(__import__, [ruffus_name]))[0]


try:
    from StringIO import StringIO
except:
    from io import StringIO


# ---------------------------------------------------------------
#   create initial files
#
@mkdir(tempdir + 'data/scratch/lg/what/one/two/three/four/five/six/seven')
@originate([[tempdir + 'data/scratch/lg/what/one/two/three/four/five/six/seven/job1.a.start', tempdir + 'job1.b.start'],
            [tempdir + 'data/scratch/lg/what/one/two/three/four/five/six/seven/job2.a.start',
                tempdir + 'job2.b.start'],
            [tempdir + 'data/scratch/lg/what/one/two/three/four/five/six/seven/job3.a.start', tempdir + 'job3.b.start']])
def create_initial_file_pairs(output_files):
    # create both files as necessary
    for output_file in output_files:
        with open(output_file, "w") as oo:
            pass

# ---------------------------------------------------------------
#   first task


@transform(create_initial_file_pairs, suffix(".start"), ".output.1")
def first_task(input_files, output_file):
    with open(output_file, "w"):
        pass


# ---------------------------------------------------------------
#   second task
@transform(first_task, suffix(".output.1"), ".output.2")
def second_task(input_files, output_file):
    with open(output_file, "w"):
        pass


test_pipeline = Pipeline("test")
test_pipeline.originate(output=[[tempdir + 'data/scratch/lg/what/one/two/three/four/five/six/seven/job1.a.start',  tempdir + 'job1.b.start'],
                                [tempdir + 'data/scratch/lg/what/one/two/three/four/five/six/seven/job2.a.start',
                                    tempdir + 'job2.b.start'],
                                [tempdir + 'data/scratch/lg/what/one/two/three/four/five/six/seven/job3.a.start', tempdir + 'job3.b.start']],
                        task_func=create_initial_file_pairs)
test_pipeline.transform(task_func=first_task, input=create_initial_file_pairs,
                        filter=suffix(".start"), output=".output.1")
test_pipeline.transform(input=first_task, filter=suffix(
    ".output.1"), output=".output.2", task_func=second_task)


decorator_syntax = 0
oop_syntax = 1


class Test_verbosity(unittest.TestCase):
    # ___________________________________________________________________________
    #
    #   test_printout_abbreviated_path1
    # ___________________________________________________________________________
    def test_printout_abbreviated_path1(self):
        """Input file exists, output doesn't exist"""
        for syntax in decorator_syntax, oop_syntax:
            s = StringIO()
            if syntax == oop_syntax:
                test_pipeline.printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=1)
            else:
                pipeline_printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=1, wrap_width=500, pipeline="main")
            ret = s.getvalue()
            self.assertTrue(
                "[[.../job1.a.start, test_verbosity/job1.b.start]" in ret)

    # ___________________________________________________________________________
    #
    #   test_printout_abbreviated_path2
    # ___________________________________________________________________________

    def test_printout_abbreviated_path2(self):
        """Input file exists, output doesn't exist"""
        for syntax in decorator_syntax, oop_syntax:
            s = StringIO()
            if syntax == oop_syntax:
                test_pipeline.printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=2, wrap_width=500)
            else:
                pipeline_printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=2, wrap_width=500, pipeline="main")
            ret = s.getvalue()
            self.assertTrue(
                "[[.../seven/job1.a.start, test_verbosity/job1.b.start]" in ret)

    # ___________________________________________________________________________
    #
    #   test_printout_abbreviated_path3
    # ___________________________________________________________________________

    def test_printout_abbreviated_path3(self):
        """Input file exists, output doesn't exist"""
        for syntax in decorator_syntax, oop_syntax:
            s = StringIO()
            if syntax == oop_syntax:
                test_pipeline.printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=3, wrap_width=500)
            else:
                pipeline_printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=3, wrap_width=500, pipeline="main")
            ret = s.getvalue()
            self.assertTrue(
                "[[.../six/seven/job1.a.start, test_verbosity/job1.b.start]" in ret)

        # ___________________________________________________________________________
    #
    #   test_printout_abbreviated_path9
    # ___________________________________________________________________________
    def test_printout_abbreviated_path9(self):
        """Input file exists, output doesn't exist"""
        for syntax in decorator_syntax, oop_syntax:
            s = StringIO()
            if syntax == oop_syntax:
                test_pipeline.printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=9, wrap_width=500)
            else:
                pipeline_printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=9, wrap_width=500, pipeline="main")
            ret = s.getvalue()
            self.assertTrue(
                "[[.../what/one/two/three/four/five/six/seven/job1.a.start, test_verbosity/job1.b.start]" in ret)

    # ___________________________________________________________________________
    #
    #   test_printout_abbreviated_path0
    # ___________________________________________________________________________

    def test_printout_abbreviated_path0(self):
        """Input file exists, output doesn't exist"""
        for syntax in decorator_syntax, oop_syntax:
            s = StringIO()
            if syntax == oop_syntax:
                test_pipeline.printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=0, wrap_width=500)
            else:
                pipeline_printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=0, wrap_width=500, pipeline="main")
            ret = s.getvalue()
            path_str = os.path.abspath(
                '%sdata/scratch/lg/what/one/two/three/four/five/six/seven/job2.a.start' % tempdir)
            path_str = '[[%s' % path_str

    # ___________________________________________________________________________
    #
    #   test_printout_abbreviated_path_minus_60
    # ___________________________________________________________________________

    def test_printout_abbreviated_path_minus_60(self):
        """Input file exists, output doesn't exist"""
        for syntax in decorator_syntax, oop_syntax:
            s = StringIO()
            if syntax == oop_syntax:
                test_pipeline.printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=-60, wrap_width=500)
            else:
                pipeline_printout(
                    s, [second_task], verbose=5, verbose_abbreviated_path=-60, wrap_width=500, pipeline="main")
            ret = s.getvalue()

            self.assertTrue(
                '[<???> ve/six/seven/job1.a.start, test_verbosity/job1.b.start]' in ret)


#
#   Necessary to protect the "entry point" of the program under windows.
#       see: http://docs.python.org/library/multiprocessing.html#multiprocessing-programming
#
if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
from __future__ import print_function
import shutil
import unittest
import json
from ruffus import transform, follows, originate, pipeline_run, Pipeline, \
    suffix, regex, mkdir, active_if, collate, merge
import sys

"""

    test_active_if.py

"""


import os
tempdir = os.path.relpath(os.path.abspath(os.path.splitext(__file__)[0])) + "/"

# add grandparent to search path for testing
grandparent_dir = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", ".."))
sys.path.insert(0, grandparent_dir)

# module name = script name without extension
module_name = os.path.splitext(os.path.basename(__file__))[0]


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   imports


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Main logic


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888


def helper(infiles, outfiles):
    if not isinstance(infiles, (tuple, list)):
        infiles = [infiles]
    if not isinstance(outfiles, list):
        outfiles = [outfiles]

    output_text = ""
    preamble_len = 0
    for infile in infiles:
        if infile:
            with open(infile) as ii:
                for line in ii:
                    output_text += line
                    preamble_len = max(preamble_len, len(
                        line) - len(line.lstrip()))

    preamble = " " * (preamble_len + 4) if len(output_text) else ""

    for outfile in outfiles:
        file_output_text = preamble + \
            json.dumps(infile) + " -> " + json.dumps(outfile) + "\n"
        with open(outfile, "w") as oo:
            oo.write(output_text + file_output_text)


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888

#   Tasks


# 88888888888888888888888888888888888888888888888888888888888888888888888888888888888888888
pipeline_active_if = True
#
#    task1
#


@follows(mkdir("test_active_if"))
@originate(['test_active_if/a.1', 'test_active_if/b.1'], "an extra_parameter")
def task1(outfile, extra):
    """
    First task
    """
    # N.B. originate works with an extra parameter
    helper(None, outfile)


#
#    task2
#
@transform(task1, suffix(".1"), ".2")
def task2(infile, outfile):
    """
    Second task
    """
    helper(infile, outfile)


#
#    task3
#
@active_if(lambda: pipeline_active_if)
@transform(task1, suffix(".1"), ".3")
def task3(infile, outfile):
    """
    Third task
    """
    helper(infile, outfile)


#
#    task4
#
@collate([task2, task3], regex(r"(.+)\.[23]"), r"\1.4")
def task4(infiles, outfile):
    """
    Fourth task
    """
    helper(infiles, outfile)

#
#    task4
#


@merge(task4, "test_active_if/summary.5")
def task5(infiles, outfile):
    """
    Fifth task
    """
    helper(infiles, outfile)


expected_active_text = """null -> "test_active_if/a.1"
    "test_active_if/a.1" -> "test_active_if/a.2"
null -> "test_active_if/a.1"
    "test_active_if/a.1" -> "test_active_if/a.3"
        "test_active_if/a.3" -> "test_active_if/a.4"
null -> "test_active_if/b.1"
    "test_active_if/b.1" -> "test_active_if/b.2"
null -> "test_active_if/b.1"
    "test_active_if/b.1" -> "test_active_if/b.3"
        "test_active_if/b.3" -> "test_active_if/b.4"
            "test_active_if/b.4" -> "test_active_if/summary.5"
"""

expected_inactive_text = """null -> "test_active_if/a.1"
    "test_active_if/a.1" -> "test_active_if/a.2"
        "test_active_if/a.2" -> "test_active_if/a.4"
null -> "test_active_if/b.1"
    "test_active_if/b.1" -> "test_active_if/b.2"
        "test_active_if/b.2" -> "test_active_if/b.4"
            "test_active_if/b.4" -> "test_active_if/summary.5"
"""


# alternative syntax
test_pipeline = Pipeline("test")
test_pipeline.originate(task1, ['test_active_if/a.1', 'test_active_if/b.1'], "an extra_parameter")\
    .follows(mkdir("test_active_if"))
test_pipeline.transform(task2, task1, suffix(".1"), ".2")
test_pipeline.transform(task3, task1, suffix(
    ".1"), ".3").active_if(lambda: pipeline_active_if)
test_pipeline.collate(task4, [task2, task3], regex(r"(.+)\.[23]"), r"\1.4")
test_pipeline.merge(task5, task4, "test_active_if/summary.5")


class Test_ruffus(unittest.TestCase):
    def setUp(self):
        try:
            shutil.rmtree(tempdir)
        except:
            pass
        os.makedirs(tempdir)

    def tearDown(self):
        try:
            shutil.rmtree(tempdir)
            pass
        except:
            pass

    def test_active_if_true(self):
        global pipeline_active_if
        pipeline_active_if = True
        pipeline_run(multiprocess=50, verbose=0, pipeline="main")

        with open("test_active_if/summary.5") as ii:
            active_text = ii.read()
        if active_text != expected_active_text:
            raise Exception("Error:\n\tExpected\n%s\nInstead\n%s\n" %
                            (active_text, expected_active_text))

    def test_active_if_false(self):
        global pipeline_active_if
        pipeline_active_if = False
        pipeline_run(multiprocess=50, verbose=0, pipeline="main")
        with open("test_active_if/summary.5") as ii:
            inactive_text = ii.read()
        if inactive_text != expected_inactive_text:
            raise Exception("Error:\n\tExpected\n%s\nInstead\n%s\n" %
                            (inactive_text, expected_inactive_text))
            shutil.rmtree("test_active_if")

    def test_newstyle_active_if_true(self):
        global pipeline_active_if
        pipeline_active_if = True
        test_pipeline.run(multiprocess=50, verbose=0)

        with open("test_active_if/summary.5") as ii:
            active_text = ii.read()
        if active_text != expected_active_text:
            raise Exception("Error:\n\tExpected\n%s\nInstead\n%s\n" %
                            (active_text, expected_active_text))

    def test_newstyle_active_if_false(self):
        global pipeline_active_if
        pipeline_active_if = False
        test_pipeline.run(multiprocess=50, verbose=0)
        with open("test_active_if/summary.5") as ii:
            inactive_text = ii.read()
        if inactive_text != expected_inactive_text:
            raise Exception("Error:\n\tExpected\n%s\nInstead\n%s\n" %
                            (inactive_text, expected_inactive_text))
            shutil.rmtree("test_active_if")


if __name__ == '__main__':
    unittest.main()
